from .animate import create_animation
from .common import (
    DEFAULT_QUANTITIES,
    PLOT_DEFAULTS,
    Quantity,
    auto_grid,
    auto_range_iqr,
)
from .convert import convert_to_xdmf
from .interactive import InteractivePlot, MultiInteractivePlot
from .io import get_plot_data, get_state_string
from .monitor import monitor_solution
from .snapshot import generate_snapshots



import logging
from typing import Any, Dict, Literal, Optional, Sequence, Tuple

import h5py
import numpy as np
from matplotlib import pyplot as plt

from ..device.device import Device
from ..solution.data import get_data_range
from .common import DEFAULT_QUANTITIES, PLOT_DEFAULTS, Quantity, auto_grid
from .io import get_plot_data, get_state_string


class InteractivePlot:
    def __init__(
        self,
        input_file: str,
        shading: Literal["flat", "gouraud"] = "gouraud",
        dimensionless: bool = False,
        xlim: Optional[Tuple[float, float]] = None,
        ylim: Optional[Tuple[float, float]] = None,
        axis_labels: bool = False,
        logger: logging.Logger = None,
    ):
        self.input_file = input_file
        self.shading = shading
        self.dimensionless = dimensionless
        self.xlim = xlim
        self.ylim = ylim
        self.axis_labels = axis_labels
        self.frame = 0
        self.quantity = Quantity.ORDER_PARAMETER
        self.logger = logger or logging.getLogger()

    def show(self):
        with h5py.File(self.input_file, "r") as h5file:
            device = Device.from_hdf5(h5file["solution/device"])
            mesh = device.mesh
            if self.dimensionless:
                scale = 1
                units_str = "\\xi"
            else:
                scale = device.layer.coherence_length
                units_str = f"{device.ureg(device.length_units).units:~L}"
            x, y = scale * mesh.sites.T
            # Get the ranges for the frame
            min_frame, max_frame = get_data_range(h5file)

            def on_keypress(event):
                if event.key == "right":
                    self.frame = np.minimum(self.frame + 1, max_frame)

                elif event.key == "left":
                    self.frame = np.maximum(self.frame - 1, min_frame)

                if event.key == "shift+right":
                    self.frame = np.minimum(self.frame + 10, max_frame)

                elif event.key == "shift+left":
                    self.frame = np.maximum(self.frame - 10, min_frame)

                elif event.key == "up":
                    self.frame = np.minimum(self.frame + 100, max_frame)

                elif event.key == "down":
                    self.frame = np.maximum(self.frame - 100, min_frame)

                elif event.key == "shift+up":
                    self.frame = np.minimum(self.frame + 1000, max_frame)

                elif event.key == "shift+down":
                    self.frame = np.maximum(self.frame - 1000, min_frame)

                elif event.key == "home":
                    self.frame = min_frame

                elif event.key == "end":
                    self.frame = max_frame

                elif event.key == "1":
                    self.quantity = Quantity.ORDER_PARAMETER

                elif event.key == "2":
                    self.quantity = Quantity.PHASE

                elif event.key == "3":
                    self.quantity = Quantity.SUPERCURRENT

                elif event.key == "4":
                    self.quantity = Quantity.NORMAL_CURRENT

                elif event.key == "5":
                    self.quantity = Quantity.SCALAR_POTENTIAL

                elif event.key == "6":
                    self.quantity = Quantity.APPLIED_VECTOR_POTENTIAL

                elif event.key == "7":
                    self.quantity = Quantity.INDUCED_VECTOR_POTENTIAL

                elif event.key == "8":
                    self.quantity = Quantity.EPSILON

                elif event.key == "9":
                    self.quantity = Quantity.VORTICITY

                draw()

            def draw():
                values, direction, limits = get_plot_data(
                    h5file, mesh, self.quantity, self.frame
                )
                state = get_state_string(h5file, self.frame, max_frame)

                fig.suptitle(f"{self.quantity.value}\n{state}")
                if self.shading == "flat":
                    # https://stackoverflow.com/questions/40492511/set-array-in-tripcolor-bug
                    values = values[mesh.elements].mean(axis=1)
                triplot.set_array(values)
                triplot.set_clim(*limits)
                triplot.set_cmap(PLOT_DEFAULTS[self.quantity].cmap)
                cbar.set_label(PLOT_DEFAULTS[self.quantity].clabel)
                fig.canvas.draw()

            # Temp data to use in plots
            temp_value = np.ones_like(x)
            temp_value[0] = 0
            temp_value[1] = 0.5

            fig, ax = plt.subplots()
            fig.subplots_adjust(top=0.8)
            fig.canvas.mpl_connect("key_press_event", on_keypress)
            triplot = ax.tripcolor(
                x, y, temp_value, triangles=mesh.elements, shading=self.shading
            )
            cbar = fig.colorbar(triplot)
            ax.set_aspect("equal")
            ax.set_xlim(self.xlim)
            ax.set_ylim(self.ylim)
            if self.axis_labels:
                ax.set_xlabel(f"$x$ [${units_str}$]")
                ax.set_ylabel(f"$y$ [${units_str}$]")
            draw()
            plt.show()


class MultiInteractivePlot:
    def __init__(
        self,
        input_file: str,
        shading: Literal["flat", "gouraud"] = "gouraud",
        dimensionless: bool = False,
        xlim: Optional[Tuple[float, float]] = None,
        ylim: Optional[Tuple[float, float]] = None,
        axis_labels: bool = False,
        quantities: Sequence[str] = DEFAULT_QUANTITIES,
        max_cols: int = 4,
        logger: logging.Logger = None,
        figure_kwargs: Optional[Dict[str, Any]] = None,
    ):
        self.input_file = input_file
        self.shading = shading
        self.dimensionless = dimensionless
        self.xlim = xlim
        self.ylim = ylim
        self.axis_labels = axis_labels
        self.frame = 0
        if quantities is None:
            quantities = Quantity.get_keys()
        self.quantities = [Quantity.from_key(name) for name in quantities]
        self.num_plots = len(quantities)
        self.max_cols = max_cols
        self.logger = logger if logger is not None else logging.getLogger()
        self.figure_kwargs = figure_kwargs or dict()
        self.figure_kwargs.setdefault("constrained_layout", True)
        default_figsize = (
            3.25 * min(self.max_cols, self.num_plots),
            3 * max(1, self.num_plots // self.max_cols),
        )
        self.figure_kwargs.setdefault("figsize", default_figsize)
        self.figure_kwargs.setdefault("sharex", True)
        self.figure_kwargs.setdefault("sharey", True)

    def show(self):
        with h5py.File(self.input_file, "r") as h5file:
            device = Device.from_hdf5(h5file["solution/device"])
            mesh = device.mesh
            if self.dimensionless:
                scale = 1
                units_str = "\\xi"
            else:
                scale = device.layer.coherence_length
                units_str = f"{device.ureg(device.length_units).units:~L}"
            x, y = scale * mesh.sites.T

            min_frame, max_frame = get_data_range(h5file)

            def on_keypress(event):
                if event.key == "right":
                    self.frame = np.minimum(self.frame + 1, max_frame)

                elif event.key == "left":
                    self.frame = np.maximum(self.frame - 1, min_frame)

                if event.key == "shift+right":
                    self.frame = np.minimum(self.frame + 10, max_frame)

                elif event.key == "shift+left":
                    self.frame = np.maximum(self.frame - 10, min_frame)

                elif event.key == "up":
                    self.frame = np.minimum(self.frame + 100, max_frame)

                elif event.key == "down":
                    self.frame = np.maximum(self.frame - 100, min_frame)

                elif event.key == "shift+up":
                    self.frame = np.minimum(self.frame + 1000, max_frame)

                elif event.key == "shift+down":
                    self.frame = np.maximum(self.frame - 1000, min_frame)

                elif event.key == "home":
                    self.frame = min_frame

                elif event.key == "end":
                    self.frame = max_frame

                draw()

            vmins = [+np.inf for _ in self.quantities]
            vmaxs = [-np.inf for _ in self.quantities]

            def draw():
                state = get_state_string(h5file, self.frame, max_frame)
                fig.suptitle(state)
                for i, (quantity, collection) in enumerate(
                    zip(self.quantities, collections)
                ):
                    values, direction, limits = get_plot_data(
                        h5file, mesh, quantity, self.frame
                    )
                    if self.shading == "flat":
                        # https://stackoverflow.com/questions/40492511/set-array-in-tripcolor-bug
                        values = values[mesh.elements].mean(axis=1)
                    collection.set_array(values)
                    vmins[i] = min(vmins[i], limits[0])
                    vmaxs[i] = max(vmaxs[i], limits[1])
                    collection.set_clim(vmins[i], vmaxs[i])
                fig.canvas.draw()

            # Temp data to use in plots
            temp_value = np.ones_like(x)
            temp_value[0] = 0
            temp_value[1] = 0.5

            fig, axes = auto_grid(
                self.num_plots, max_cols=self.max_cols, **self.figure_kwargs
            )
            fig.canvas.mpl_connect("key_press_event", on_keypress)

            collections = []
            for quantity, ax in zip(self.quantities, axes.flat):
                ax: plt.Axes
                opts = PLOT_DEFAULTS[quantity]
                collection = ax.tripcolor(
                    x,
                    y,
                    temp_value,
                    triangles=mesh.elements,
                    shading=self.shading,
                    cmap=opts.cmap,
                )
                cbar = fig.colorbar(collection, ax=ax)
                cbar.set_label(opts.clabel)
                ax.set_aspect("equal")
                ax.set_title(quantity.value)
                collections.append(collection)
                ax.set_xlim(self.xlim)
                ax.set_ylim(self.ylim)
                if self.axis_labels:
                    ax.set_xlabel(f"$x$ [${units_str}$]")
                    ax.set_ylabel(f"$y$ [${units_str}$]")

            draw()
            plt.show()



import inspect
import itertools
import logging
import math
import os
from datetime import datetime
from typing import Callable, Dict, List, NamedTuple, Optional, Sequence, Tuple, Union

import numpy as np
import scipy.sparse as sp

try:
    import cupy  # type: ignore
except ImportError:
    cupy = None

try:
    import pypardiso  # type: ignore
except ImportError:
    pypardiso = None

from ..device.device import Device, TerminalInfo
from ..finite_volume.operators import MeshOperators
from ..parameter import Parameter
from ..solution.solution import Solution
from ..sources.constant import ConstantField
from .options import SolverOptions, SparseSolver
from .runner import DataHandler, Runner, RunningState
from .screening import get_A_induced_cupy, get_A_induced_numba

logger = logging.getLogger("solver")


def validate_terminal_currents(
    terminal_currents: Union[Callable, Dict[str, float]],
    terminal_info: Sequence[TerminalInfo],
    solver_options: SolverOptions,
    num_evals: int = 100,
) -> None:
    """Ensure that the terminal currents satisfy current conservation."""

    def check_total_current(currents: Dict[str, float]):
        names = set([t.name for t in terminal_info])
        if unknown := set(currents).difference(names):
            raise ValueError(
                f"Unknown terminal(s) in terminal currents: {list(unknown)}."
            )
        total_current = sum(currents.values())
        if total_current:
            raise ValueError(
                f"The sum of all terminal currents must be 0 (got {total_current:.2e})."
            )

    if callable(terminal_currents):
        times = np.random.default_rng().random(num_evals) * solver_options.solve_time
        for t in times:
            check_total_current(terminal_currents(t))
    else:
        check_total_current(terminal_currents)


class SolverResult(NamedTuple):
    """A container for the results of a single solve step.

    dt: The time step size used for the solve step
    psi: The order parameter
    mu: The scalar potential
    supercurrent: The supercurrent density
    normal_current: The normal current density
    A_induced: The induced vector potential
    A_applied: The applied vector potential. This will be ``None`` in the case of
        a time-independent vector potential.
    epsilon: The disorder parameter, ``epsilon``. This will be ``None`` in the case of
        a time-independent ``epsilon``.
    """

    dt: float
    psi: np.ndarray
    mu: np.ndarray
    supercurrent: np.ndarray
    normal_current: np.ndarray
    A_induced: np.ndarray
    A_applied: Optional[np.ndarray] = None
    epsilon: Optional[np.ndarray] = None


class TDGLSolver:
    """Solver for a TDGL model.

    An instance of :class:`tdgl.TDGLSolver` is created and executed
    by calling :func:`tdgl.solve`.

    Args:
        device: The :class:`tdgl.Device` to solve.
        options: An instance :class:`tdgl.SolverOptions` specifying the solver
            parameters.
        applied_vector_potential: A function or :class:`tdgl.Parameter` that computes
            the applied vector potential as a function of position ``(x, y, z)``,
            or of position and time ``(x, y, z, *, t)``. If a float ``B`` is given,
            the applied vector potential will be that of a uniform magnetic field with
            strength ``B`` ``field_units``. If the applied vector potential is time-dependent,
            this argument must be a :class:`tdgl.Parameter`.
        terminal_currents: A dict of ``{terminal_name: current}`` or a callable with signature
            ``func(time: float) -> {terminal_name: current}``, where ``current`` is a float
            in units of ``current_units`` and ``time`` is the dimensionless time.
        disorder_epsilon: A float in range [-1, 1], or a function that returns
            :math:`\\epsilon\\in[-1, 1]` as a function of position ``r=(x, y)`` or
            position and time ``(x, y, *, t)``.
            Setting :math:`\\epsilon(\\mathbf{r}, t)=T_c/T - 1 < 1` suppresses the
            order parameter at position :math:`\\mathbf{r}=(x, y)`, which can be used
            to model inhomogeneity.
        seed_solution: A :class:`tdgl.Solution` instance to use as the initial state
            for the simulation.
    """

    def __init__(
        self,
        device: Device,
        options: SolverOptions,
        applied_vector_potential: Union[Callable, float] = 0.0,
        terminal_currents: Union[Callable, Dict[str, float], None] = None,
        disorder_epsilon: Union[Callable, float] = 1.0,
        seed_solution: Optional[Solution] = None,
    ):
        self.device = device
        self.options = options
        self.options.validate()
        self.terminal_currents = terminal_currents
        self.seed_solution = seed_solution

        if self.options.gpu:
            assert cupy is not None
            self.xp = cupy
            self.use_cupy = True
        else:
            self.xp = np
            self.use_cupy = False

        mesh = self.device.mesh
        ureg = self.device.ureg
        self.probe_points = device.probe_point_indices
        length_units = ureg(self.device.length_units)
        field_units = options.field_units
        current_units = options.current_units

        edges = mesh.edge_mesh.edges
        self.num_edges = len(edges)
        normalized_directions = mesh.edge_mesh.normalized_directions
        length_units = ureg(device.length_units)
        xi = device.coherence_length.magnitude
        self.u = device.layer.u
        self.gamma = device.layer.gamma
        K0 = device.K0
        A0 = device.A0
        Bc2 = device.Bc2

        # The vector potential is evaluated on the mesh edges,
        # where the edge coordinates are in dimensionful units.
        self.sites = xi * mesh.sites
        self.edge_centers = xi * mesh.edge_mesh.centers
        self.z0 = device.layer.z0 * np.ones(len(self.edge_centers), dtype=float)

        self.dynamic_vector_potential = (
            isinstance(applied_vector_potential, Parameter)
            and applied_vector_potential.time_dependent
        )
        if not callable(applied_vector_potential):
            applied_vector_potential = ConstantField(
                applied_vector_potential,
                field_units=field_units,
                length_units=device.length_units,
            )
        self.applied_vector_potential = applied_vector_potential
        # Evaluate the vector potential
        self.A_scale = (
            (ureg(field_units) * length_units / (Bc2 * xi * length_units))
            .to_base_units()
            .magnitude
        )
        A_kwargs = dict(t=0) if self.dynamic_vector_potential else dict()
        current_A_applied = self.applied_vector_potential(
            self.edge_centers[:, 0], self.edge_centers[:, 1], self.z0, **A_kwargs
        )
        current_A_applied = self.A_scale * np.asarray(current_A_applied)[:, :2]
        if current_A_applied.shape != self.edge_centers.shape:
            raise ValueError(
                f"Unexpected shape for vector_potential: {current_A_applied.shape}."
            )

        # Create the epsilon parameter, which sets the local critical temperature.
        if callable(disorder_epsilon):
            argspec = inspect.getfullargspec(disorder_epsilon)
            self.dynamic_epsilon = "t" in argspec.kwonlyargs
            self.vectorized_epsilon = (
                argspec.kwonlydefaults is not None
                and argspec.kwonlydefaults.get("vectorized", False)
            )
        else:
            # epsilon constant as a function of both position and time
            _disorder_epsilon = disorder_epsilon

            def disorder_epsilon(r):
                return _disorder_epsilon * np.ones(len(r), dtype=float)

            self.vectorized_epsilon = True
            self.dynamic_epsilon = False

        self.disorder_epsilon = disorder_epsilon
        kw = dict(t=0) if self.dynamic_epsilon else dict()
        if self.vectorized_epsilon:
            epsilon = disorder_epsilon(self.sites, **kw)
        else:
            epsilon = np.array([float(disorder_epsilon(r, **kw)) for r in self.sites])
        if np.any(epsilon < -1) or np.any(epsilon > 1):
            raise ValueError("The disorder parameter epsilon must be in range [-1, 1].")

        # Find the current terminal sites.
        self.terminal_info = device.terminal_info()
        self.terminal_names = [term.name for term in self.terminal_info]
        for term_info in self.terminal_info:
            if term_info.length == 0:
                raise ValueError(
                    f"Terminal {term_info.name!r} does not contain any points"
                    " on the boundary of the mesh."
                )
        # Define the source-drain current.
        if terminal_currents and device.probe_points is None:
            logger.warning(
                "The terminal currents are non-null, but the device has no probe points."
            )
        terminal_names = [term.name for term in self.terminal_info]
        if terminal_currents is None:
            terminal_currents = {name: 0 for name in terminal_names}
        if callable(terminal_currents):
            current_func = terminal_currents
        else:
            terminal_currents = {
                name: terminal_currents.get(name, 0) for name in terminal_names
            }

            def current_func(t):
                return terminal_currents

        J_scale = 4 * ((ureg(current_units) / length_units) / K0).to_base_units()
        assert "dimensionless" in str(J_scale.units), str(J_scale.units)
        J_scale = J_scale.magnitude
        self.current_func = lambda t: {
            key: J_scale * value for key, value in current_func(t).items()
        }
        validate_terminal_currents(self.current_func, self.terminal_info, self.options)
        terminal_indices = [t.site_indices for t in self.terminal_info]
        if terminal_indices:
            normal_boundary_index = np.concatenate(terminal_indices, dtype=np.int64)
        else:
            normal_boundary_index = np.array([], dtype=np.int64)
        # Cache the terminal current densities at each time step.
        # Only update the mu boundary conditions if the current has changed.
        self.terminal_current_densities = {name: 0 for name in self.terminal_names}

        # Construct finite-volume operators
        terminal_psi = options.terminal_psi
        logger.info("Constructing finite volume operators.")
        operators = MeshOperators(
            mesh,
            options.sparse_solver,
            use_cupy=self.use_cupy,
            fixed_sites=normal_boundary_index,
            fix_psi=(terminal_psi is not None),
        )
        operators.build_operators()
        operators.set_link_exponents(current_A_applied)
        self.operators = operators
        if options.sparse_solver is SparseSolver.PARDISO:
            assert self.operators.mu_laplacian_lu is None
            assert pypardiso is not None

        # Initialize the order parameter and electric potential
        psi_init = np.ones(len(mesh.sites), dtype=np.complex128)
        if terminal_psi is not None:
            psi_init[normal_boundary_index] = terminal_psi
        mu_init = np.zeros(len(mesh.sites))
        mu_boundary = np.zeros_like(mesh.edge_mesh.boundary_edge_indices, dtype=float)

        if self.use_cupy:
            epsilon = cupy.asarray(epsilon)
            mu_boundary = cupy.asarray(mu_boundary)
            normalized_directions = cupy.asarray(normalized_directions)
            current_A_applied = cupy.asarray(current_A_applied)

        self.psi_init = psi_init
        self.mu_init = mu_init
        self.epsilon = epsilon
        self.mu_boundary = mu_boundary
        self.normalized_directions = normalized_directions
        self.current_A_applied = current_A_applied

        self.new_A_induced = None
        self.areas = None
        if options.include_screening:
            A_scale = (ureg("mu_0") / (4 * np.pi) * K0 / A0).to(1 / length_units)
            self.new_A_induced = np.empty((self.num_edges, 2), dtype=float)
            self.areas = A_scale.magnitude * mesh.areas * xi**2
            if self.use_cupy:
                self.areas = cupy.asarray(self.areas)
                self.edge_centers = cupy.asarray(self.edge_centers)
                self.sites = cupy.asarray(self.sites)
                self.new_A_induced = cupy.asarray(self.new_A_induced)

        # Running list of the max abs change in |psi|^2 between subsequent solve steps.
        # This list is used to calculate the adaptive time step.
        self.d_psi_sq_vals = []
        self.tentative_dt = options.dt_init
        self.dt_max = options.dt_max if options.adaptive else options.dt_init

        if options.monitor:
            os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"

    def update_mu_boundary(self, time: float) -> None:
        """Computes the terminal current density for a given time step and
        updates the scalar potential boundary conditions accordingly.

        Args:
            time: The current value of the dimensionless time.
        """
        # Compute the current density for this step
        # and update the current boundary conditions.
        currents = self.current_func(time)
        terminal_current_densities = self.terminal_current_densities
        for terminal in self.terminal_info:
            current_density = (-1 / terminal.length) * sum(
                currents.get(name, 0)
                for name in self.terminal_names
                if name != terminal.name
            )
            # Only update mu_boundary if the terminal current has changed
            if current_density != terminal_current_densities[terminal.name]:
                terminal_current_densities[terminal.name] = current_density
                self.mu_boundary[terminal.boundary_edge_indices] = current_density

    def update_applied_vector_potential(self, time: float) -> np.ndarray:
        """Evaluates the time-dependent vector potential.

        Args:
            time: The current value of the dimensionless time.

        Returns:
            The new value of the applied vector potential.
        """
        A_applied = self.applied_vector_potential(
            self.edge_centers[:, 0], self.edge_centers[:, 1], self.z0, t=time
        )
        A_applied = self.A_scale * A_applied[:, :2]
        if self.use_cupy:
            A_applied = cupy.asarray(A_applied)
        return A_applied

    def update_epsilon(self, time: float) -> np.ndarray:
        """Evaluates the time-dependent disorder parameter :math:`\\epsilon`.

        Args:
            time: The current value of the dimensionless time.

        Returns:
            The new value of :math:`\\epsilon`
        """
        if self.vectorized_epsilon:
            epsilon = self.disorder_epsilon(self.sites, t=time)
        else:
            epsilon = np.array(
                [float(self.disorder_epsilon(r, t=time)) for r in self.sites]
            )
        if self.use_cupy:
            epsilon = cupy.asarray(epsilon)
        return epsilon

    @staticmethod
    def solve_for_psi_squared(
        *,
        psi: np.ndarray,
        abs_sq_psi: np.ndarray,
        mu: np.ndarray,
        epsilon: np.ndarray,
        gamma: float,
        u: float,
        dt: float,
        psi_laplacian: sp.spmatrix,
    ) -> Union[Tuple[np.ndarray, np.ndarray], None]:
        """Solves for :math:`\\psi^{n+1}` and :math:`|\\psi^{n+1}|^2` given
        :math:`\\psi^n` and :math:`\\mu^n`.

        Args:
            psi: The current value of the order parameter, :math:`\\psi^n`
            abs_sq_psi: The current value of the superfluid density, :math:`|\\psi^n|^2`
            mu: The current value of the electric potential, :math:`\\mu^n`
            epsilon: The disorder parameter, :math:`\\epsilon`
            gamma: The inelastic scattering parameter, :math:`\\gamma`.
            u: The ratio of relaxation times for the order parameter, :math:`u`
            dt: The time step
            psi_laplacian: The covariant Laplacian for the order parameter

        Returns:
            ``None`` if the calculation failed to converge, otherwise the new order
            parameter :math:`\\psi^{n+1}` and superfluid density :math:`|\\psi^{n+1}|^2`.
        """
        if isinstance(psi, np.ndarray):
            xp = np
        else:
            assert cupy is not None
            assert isinstance(psi, cupy.ndarray)
            xp = cupy
        U = xp.exp(-1j * mu * dt)
        z = U * gamma**2 / 2 * psi
        with np.errstate(all="raise"):
            try:
                w = z * abs_sq_psi + U * (
                    psi
                    + (dt / u)
                    * xp.sqrt(1 + gamma**2 * abs_sq_psi)
                    * ((epsilon - abs_sq_psi) * psi + psi_laplacian @ psi)
                )
                c = w.real * z.real + w.imag * z.imag
                two_c_1 = 2 * c + 1
                w2 = xp.absolute(w) ** 2
                discriminant = two_c_1**2 - 4 * xp.absolute(z) ** 2 * w2
            except Exception:
                logger.warning("Unable to solve for |psi|^2.", exc_info=True)
                return None
        if xp.any(discriminant < 0):
            return None
        new_sq_psi = (2 * w2) / (two_c_1 + xp.sqrt(discriminant))
        psi = w - z * new_sq_psi
        return psi, new_sq_psi

    def adaptive_euler_step(
        self,
        step: int,
        psi: np.ndarray,
        abs_sq_psi: np.ndarray,
        mu: np.ndarray,
        epsilon: np.ndarray,
        dt: float,
    ) -> Tuple[np.ndarray, np.ndarray, float]:
        """Updates the order parameter and time step in an adaptive Euler step.

        Args:
            step: The solve step index, :math:`n`
            psi: The current value of the order parameter, :math:`\\psi^n`
            abs_sq_psi: The current value of the superfluid density, :math:`|\\psi^n|^2`
            mu: The current value of the electric potential, :math:`\\mu^n`
            epsilon: The disorder parameter :math:`\\epsilon^n`
            dt: The tentative time step, which will be updated

        Returns:
            :math:`\\psi^{n+1}`, :math:`|\\psi^{n+1}|^2`, and :math:`\\Delta t^{n}`.
        """
        options = self.options
        kwargs = dict(
            psi=psi,
            abs_sq_psi=abs_sq_psi,
            mu=mu,
            epsilon=epsilon,
            gamma=self.gamma,
            u=self.u,
            dt=dt,
            psi_laplacian=self.operators.psi_laplacian,
        )
        result = self.solve_for_psi_squared(**kwargs)
        for retries in itertools.count():
            if result is not None:
                break  # First evaluation of |psi|^2 was successful.
            if not options.adaptive or retries > options.max_solve_retries:
                raise RuntimeError(
                    f"Solver failed to converge in {options.max_solve_retries}"
                    f" retries at step {step} with dt = {dt:.2e}."
                    f" Try using a smaller dt_init."
                )
            kwargs["dt"] = dt = dt * options.adaptive_time_step_multiplier
            result = self.solve_for_psi_squared(**kwargs)
        psi, new_sq_psi = result
        return psi, new_sq_psi, dt

    def solve_for_observables(
        self, psi: np.ndarray, dA_dt: Union[float, np.ndarray]
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Solves for the scalar potential :math:`\\mu`, the supercurrent density
        :math:`\\mathbf{J}_s`, and the normal current density :math:`\\mathbf{J}_n`.

        Args:
            psi: The order parameter.
            dA_dt: The time-derivative of the vector potential.

        Returns:
            :math:`\\mu`, :math:`\\mathbf{J}_s`, and :math:`\\mathbf{J}_n`
        """
        use_cupy = self.use_cupy
        options = self.options
        use_cupy_solver = options.sparse_solver is SparseSolver.CUPY
        operators = self.operators
        # Compute the supercurrent, scalar potential, and normal current
        supercurrent = operators.get_supercurrent(psi)
        rhs = (operators.divergence @ (supercurrent - dA_dt)) - (
            operators.mu_boundary_laplacian @ self.mu_boundary
        )
        if use_cupy and not use_cupy_solver:
            rhs = cupy.asnumpy(rhs)
        if self.options.sparse_solver is SparseSolver.PARDISO:
            mu = pypardiso.spsolve(operators.mu_laplacian, rhs)
        else:
            mu = operators.mu_laplacian_lu(rhs)
        if use_cupy and not use_cupy_solver:
            mu = cupy.asarray(mu)
        normal_current = -(operators.mu_gradient @ mu) - dA_dt
        return mu, supercurrent, normal_current

    def get_induced_vector_potential(
        self,
        current_density: np.ndarray,
        A_induced_vals: List[np.ndarray],
        velocity: List[np.ndarray],
    ) -> Tuple[np.ndarray, float]:
        """Computes a new value of the induced vector potential based on Polyak's method.

        Args:
            current_density: The total current density :math:`\\mathbf{J}_s + \\mathbf{J}_n`
            A_induced_vals: A running list of the induced vector potential for previous
                iterations of Polyak's method.
            velocity: A running list of the "velocities" for previous iterations of
                Polyak's method.

        Returns:
            A new value for the induced vector potential, and the relative error in the
            induced vector potential between this iteration of Polyak's method and the
            previous iteration.
        """
        xp = self.xp
        use_cupy = self.use_cupy
        options = self.options
        mesh = self.device.mesh
        alpha = options.screening_step_size
        beta = options.screening_step_drag
        # Evaluate the induced vector potential.
        J_site = mesh.get_quantity_on_site(current_density, use_cupy=use_cupy)
        areas = self.areas
        sites = self.sites
        edge_centers = self.edge_centers
        if use_cupy:
            threads_per_block = 512
            num_blocks = math.ceil(self.num_edges / threads_per_block)
            get_A_induced_cupy(
                (num_blocks,),
                (threads_per_block, 2),
                (J_site, areas, sites, edge_centers, self.new_A_induced),
            )
        else:
            get_A_induced_numba(J_site, areas, sites, edge_centers, self.new_A_induced)
        new_A_induced = self.new_A_induced
        # Update induced vector potential using Polyak's method
        A_induced = A_induced_vals[-1]
        dA = new_A_induced - A_induced
        velocity.append((1 - beta) * velocity[-1] + alpha * dA)
        A_induced = A_induced + velocity[-1]
        A_induced_vals.append(A_induced)
        if len(A_induced_vals) > 1:
            numerator = xp.linalg.norm(dA, axis=1)
            denominator = xp.linalg.norm(A_induced, axis=1)
            # Avoid division by zero in the case of zero A_induced
            denominator = xp.maximum(denominator, 1e-20, out=denominator)
            screening_error = float(xp.max(numerator / denominator))
            del velocity[:-2]
            del A_induced_vals[:-2]
        return A_induced, screening_error

    def update(
        self,
        state: Dict[str, Union[int, float]],
        running_state: RunningState,
        dt: float,
        *,
        psi: np.ndarray,
        mu: np.ndarray,
        supercurrent: np.ndarray,
        normal_current: np.ndarray,
        induced_vector_potential: np.ndarray,
        applied_vector_potential: Optional[np.ndarray] = None,
        epsilon: Optional[np.ndarray] = None,
    ) -> SolverResult:
        """This method is called at each time step to update the state of the system.

        Args:
            state: The solver state, i.e., the solve step, time, and time step
            running_state: A container for scalar data that is saved at each time step
            dt: The time step for the previous solve step
            psi: The order parameter
            mu: The scalar potential
            supercurrent: The supercurrent density
            normal_current: The normal current density
            induced_vector_potential: The induced vector potential
            applied_vector_potential: The applied vector potential. This will be ``None``
                in the case of a time-independent vector potential.
            epsilon: The disorder parameter ``epsilon``. This will be ``None``
                in the case of a time-independent ``epsilon``.

        Returns:
            A :class:`tdgl.SolverResult` instance for the solve step.
        """
        xp = self.xp
        options = self.options
        operators = self.operators

        step = state["step"]
        time = state["time"]
        A_induced = induced_vector_potential
        prev_A_applied = A_applied = applied_vector_potential

        # Update the scalar potential boundary conditions.
        self.update_mu_boundary(time)

        # Update the applied vector potential.
        dA_dt = 0.0
        current_A_applied = self.current_A_applied
        if self.dynamic_vector_potential:
            current_A_applied = self.update_applied_vector_potential(time)
            dA_dt = xp.einsum(
                "ij, ij -> i",
                (current_A_applied - prev_A_applied) / dt,
                self.normalized_directions,
            )
            if xp.any(xp.absolute(dA_dt) > 0):
                # Update the link exponents only if the applied vector potential
                # has actually changed.
                operators.set_link_exponents(current_A_applied)
        else:
            assert A_applied is None
            prev_A_applied = A_applied = current_A_applied
        self.current_A_applied = current_A_applied

        # Update the value of epsilon
        epsilon = self.epsilon
        if self.dynamic_epsilon:
            epsilon = self.epsilon = self.update_epsilon(time)

        old_sq_psi = xp.absolute(psi) ** 2
        screening_error = np.inf
        A_induced_vals = [A_induced]
        velocity = [0.0]  # Velocity for Polyak's method
        # This loop runs only once if options.include_screening is False
        for screening_iteration in itertools.count():
            if screening_error < options.screening_tolerance:
                break
            if screening_iteration > options.max_iterations_per_step:
                raise RuntimeError(
                    f"Screening calculation failed to converge at step {step} after"
                    f" {options.max_iterations_per_step} iterations. Relative error in"
                    f" induced vector potential: {screening_error:.2e}"
                    f" (tolerance: {options.screening_tolerance:.2e})."
                )

            # Adjust the time step and calculate the new the order parameter
            if screening_iteration == 0:
                # Find a new time step only for the first screening iteration.
                dt = self.tentative_dt

            if options.include_screening:
                # Update the link variables in the covariant Laplacian and gradient
                # for psi based on the induced vector potential from the previous iteration.
                operators.set_link_exponents(current_A_applied + A_induced)

            # Update the order parameter using an adaptive time step
            psi, abs_sq_psi, dt = self.adaptive_euler_step(
                step, psi, old_sq_psi, mu, epsilon, dt
            )
            # Update the scalar potential, supercurrent density, and normal current density
            mu, supercurrent, normal_current = self.solve_for_observables(psi, dA_dt)

            if options.include_screening:
                # Evaluate the induced vector potential
                A_induced, screening_error = self.get_induced_vector_potential(
                    supercurrent + normal_current, A_induced_vals, velocity
                )
            else:
                break

        running_state.append("dt", dt)
        if self.probe_points is not None:
            # Update the voltage and phase difference
            running_state.append("mu", mu[self.probe_points])
            running_state.append("theta", xp.angle(psi[self.probe_points]))
        if options.include_screening:
            running_state.append("screening_iterations", screening_iteration)

        if options.adaptive:
            # Compute the max abs change in |psi|^2, averaged over the adaptive window,
            # and use it to select a new time step.
            self.d_psi_sq_vals.append(float(xp.absolute(abs_sq_psi - old_sq_psi).max()))
            window = options.adaptive_window
            if step > window:
                new_dt = options.dt_init / max(
                    1e-10, np.mean(self.d_psi_sq_vals[-window:])
                )
                self.tentative_dt = np.clip(0.5 * (new_dt + dt), 0, self.dt_max)

        results = [dt, psi, mu, supercurrent, normal_current, A_induced]
        if self.dynamic_vector_potential:
            results.append(current_A_applied)
        if self.dynamic_epsilon:
            results.append(epsilon)
        return SolverResult(*results)

    def solve(self) -> Optional[Solution]:
        """Runs the solver.

        Returns:
            A :class:`tdgl.Solution` instance. Returns ``None`` if the simulation was
            cancelled during the thermalization stage.
        """
        start_time = datetime.now()
        options = self.options
        options.validate()
        output_file = options.output_file
        seed_solution = self.seed_solution
        num_edges = self.num_edges
        probe_points = self.probe_points

        # Set the initial conditions.
        if self.seed_solution is None:
            parameters = {
                "psi": self.psi_init,
                "mu": self.mu_init,
                "supercurrent": np.zeros(num_edges),
                "normal_current": np.zeros(num_edges),
                "induced_vector_potential": np.zeros((num_edges, 2)),
            }
        else:
            if self.seed_solution.device != self.device:
                raise ValueError(
                    "The seed_solution.device must be equal to the device being simulated."
                )
            seed_data = seed_solution.tdgl_data
            parameters = {
                "psi": seed_data.psi,
                "mu": seed_data.mu,
                "supercurrent": seed_data.supercurrent,
                "normal_current": seed_data.normal_current,
                "induced_vector_potential": seed_data.induced_vector_potential,
            }

        fixed_values = []
        fixed_names = []
        if self.dynamic_vector_potential:
            parameters["applied_vector_potential"] = self.current_A_applied
        else:
            fixed_values.append(self.current_A_applied)
            fixed_names.append("applied_vector_potential")
        if self.dynamic_epsilon:
            parameters["epsilon"] = self.epsilon
        else:
            fixed_values.append(self.epsilon)
            fixed_names.append("epsilon")

        if self.use_cupy:
            # Move arrays to the GPU
            for key, val in parameters.items():
                parameters[key] = cupy.asarray(val)
            fixed_values = tuple(cupy.asarray(val) for val in fixed_values)

        running_names_and_sizes = {"dt": 1}
        if probe_points is not None:
            running_names_and_sizes["mu"] = len(probe_points)
            running_names_and_sizes["theta"] = len(probe_points)
        if options.include_screening:
            running_names_and_sizes["screening_iterations"] = 1

        with DataHandler(output_file=output_file, logger=logger) as data_handler:
            data_handler.save_mesh(self.device.mesh)
            if data_handler.tmp_file is not None:
                self.device.to_hdf5(
                    data_handler.tmp_file.create_group("solution/device")
                )
            logger.info(
                f"Simulation started at {start_time}"
                f" using sparse solver {options.sparse_solver.value!r}"
                f" and backend {('CuPy' if self.use_cupy else 'NumPy')!r}."
            )
            runner = Runner(
                function=self.update,
                options=options,
                data_handler=data_handler,
                monitor=options.monitor,
                monitor_update_interval=options.monitor_update_interval,
                initial_values=list(parameters.values()),
                names=list(parameters),
                fixed_values=tuple(fixed_values),
                fixed_names=tuple(fixed_names),
                running_names_and_sizes=running_names_and_sizes,
                logger=logger,
            )
            data_was_generated = runner.run()
            end_time = datetime.now()
            logger.info(f"Simulation ended at {end_time}")
            logger.info(f"Simulation took {end_time - start_time}")

            solution = None
            if data_was_generated:
                solution = Solution(
                    device=self.device,
                    path=data_handler.output_path,
                    options=options,
                    applied_vector_potential=self.applied_vector_potential,
                    terminal_currents=self.terminal_currents,
                    disorder_epsilon=self.disorder_epsilon,
                    total_seconds=(end_time - start_time).total_seconds(),
                )
                solution.to_hdf5()
            return solution



import numpy as np
import pytest
import scipy.spatial.distance

import tdgl


@pytest.mark.parametrize("metric", ("euclidean", "sqeuclidean"))
def test_cdist_invalid_shape(metric):
    XA = np.random.random((100, 4))

    XB = np.random.random((333, 4))
    with pytest.raises(ValueError):
        tdgl.distance.cdist(XA, XB, metric=metric)

    XB = np.random.random((333, 2))
    with pytest.raises(ValueError):
        tdgl.distance.cdist(XA, XB, metric=metric)


def test_cdist_invalid_metric():
    XA = np.random.random((100, 2))
    XB = np.random.random((333, 2))

    with pytest.raises(ValueError):
        tdgl.distance.cdist(XA, XB, metric="invalid")


@pytest.mark.parametrize("metric", ("euclidean", "sqeuclidean"))
@pytest.mark.parametrize("dtype", ("float64", "float32"))
@pytest.mark.parametrize("ndim", (2, 3))
def test_cdist(metric, dtype, ndim):
    XA = np.random.random((100, ndim)).astype(dtype)
    XB = np.random.random((333, ndim)).astype(dtype)
    dist_tdgl = tdgl.distance.cdist(XA, XB, metric=metric)
    dist_sp = scipy.spatial.distance.cdist(XA, XB, metric=metric)
    assert np.allclose(dist_tdgl, dist_sp)



from typing import Optional

import h5py
from tqdm import tqdm

from ..solution.solution import Solution
from .common import Quantity
from .io import get_plot_data


def convert_to_xdmf(
    path_to_solution: str,
    xdmf_path: Optional[str] = None,
    dimensionless: bool = False,
) -> None:
    """Convert a :class:`tdgl.Solution` from `HDF5 <https://www.hdfgroup.org/solutions/hdf5/>`_
    to `XDMF <https://www.xdmf.org/index.php/Main_Page>`_.

    XDMF files can be viewed using tools like `ParaView <https://www.paraview.org/>`_.
    This function requires the ``meshio`` Python package.

    Args:
        path_to_solution: Path to the HDF5 file containing the tdgl.Solution
        xdmf_path: Path to the output XDMF file.
            Defaults to ``path_to_solution.replace('.h5', '-converted.xdmf')``
        dimensionless: Save the mesh in dimensionless units
            (scaled to the coherence length).
    """
    try:
        import meshio
    except ImportError as e:
        raise RuntimeError(
            "convert_to_xdmf() requires the meshio Python package."
        ) from e

    solution = Solution.from_hdf5(path_to_solution)
    device = solution.device
    mesh = device.mesh
    points = mesh.sites
    if not dimensionless:
        points = points * device.layer.coherence_length
    cells = [("triangle", mesh.elements)]

    if xdmf_path is None:
        xdmf_path = path_to_solution.replace(".h5", "-converted.xdmf")

    first, last = solution.data_range

    with h5py.File(path_to_solution, "r") as h5file:
        with meshio.xdmf.TimeSeriesWriter(xdmf_path) as writer:
            writer.write_points_cells(points, cells)
            for step in tqdm(range(first, last + 1), desc="Solve steps"):
                solution.solve_step = step
                t = solution.tdgl_data.state["time"]
                point_data = {}
                for name, quantity in Quantity.__members__.items():
                    data, _, _ = get_plot_data(h5file, mesh, quantity, step)
                    point_data[name.lower()] = data
                writer.write_data(t, point_data=point_data)



import dataclasses
from typing import Any, Dict, List, Literal, Optional, Sequence, Tuple, Union

import h5py
import matplotlib.pyplot as plt
import matplotlib.tri as mtri
import numpy as np
from tqdm import tqdm

from ..finite_volume.mesh import Mesh
from ..geometry import path_vectors


def get_data_range(h5file: h5py.File) -> Tuple[int, int]:
    """Returns the minimum and maximum solve steps in the file."""
    keys = np.asarray([int(key) for key in h5file["data"]])
    return np.min(keys), np.max(keys)


def load_state_data(h5file: h5py.File, step: int) -> Dict[str, Any]:
    """Returns a dict of state data for the given solve step."""
    return dict(h5file["data"][str(step)].attrs)


def array_safe_equals(a: Any, b: Any) -> bool:
    """Check if a and b are equal, even if they are numpy arrays."""
    if a is b:
        return True
    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        return a.shape == b.shape and np.allclose(a, b)
    try:
        return a == b
    except TypeError:
        return NotImplemented


def dataclass_equals(dc1: Any, dc2: Any) -> bool:
    """Check if two dataclasses that may hold numpy arrays are equal."""
    if dc1 is dc2:
        return True
    if dc1.__class__ is not dc2.__class__:
        return NotImplemented
    t1 = dataclasses.astuple(dc1)
    t2 = dataclasses.astuple(dc2)
    return all(array_safe_equals(a1, a2) for a1, a2 in zip(t1, t2))


def get_edge_quantity_data(
    quantity_on_edges: np.ndarray, mesh: Mesh
) -> Tuple[np.ndarray, np.ndarray, Tuple[float, float]]:
    """Returns the value of a vector quantity living on the edges of the mesh,
    evaluated at sites in the mesh.

    Args:
        quantity_on_edges: An array of quantity values on the mesh edges.
        mesh: The :class:`tdgl.finite_volume.mesh.Mesh` instance.

    Returns:
        The magnitude and directions of the quantity on the sites, and a tuple
        of the (min, max) of the data.
    """
    directions = mesh.get_quantity_on_site(quantity_on_edges)
    norm = np.linalg.norm(directions, axis=1)
    directions /= np.maximum(norm, 1e-12)[:, np.newaxis]
    return norm, directions, (np.min(norm), np.max(norm))


@dataclasses.dataclass(eq=False)
class TDGLData:
    """A container for raw data from the TDGL solver at a single solve step.

    Args:
        step: The solver iteration.
        epsilon: The disorder parameter. :math:`\\epsilon<1` weakens the
            order parameter.
        psi: The complex order parameter at each site in the mesh.
        mu: The scalar potential at each site in the mesh.
        applied_vector_potential: The applied vector potential at each edge in the mesh.
        induced_vector_potential: The induced vector potential at each edge in the mesh.
        supercurrent: The supercurrent density at each edge in the mesh.
        normal_current: The normal density at each edge in the mesh.
        state: The solver state for the current iteration.
    """

    step: int
    epsilon: np.ndarray
    psi: np.ndarray
    mu: np.ndarray
    applied_vector_potential: np.ndarray
    induced_vector_potential: np.ndarray
    supercurrent: np.ndarray
    normal_current: np.ndarray
    state: Dict[str, Any]

    @staticmethod
    def from_hdf5(h5file: Union[h5py.File, h5py.Group], step: int) -> "TDGLData":
        """Load a :class:`TDGLData` instance from an output :class:`h5py.File`.

        Args:
            h5file: An open HDF5 output file.
            step: The solver iteration for which to load data.

        Returns:
            A :class:`TDGLData` instance containing data for the requested solve step.
        """
        step = str(step)

        def get(key, default=None):
            if key in ["step"]:
                return int(step)
            if key in ["state"]:
                return load_state_data(h5file, step)
            if key in h5file:
                dset = h5file[key]
                dset.refresh()
                return np.asarray(dset)
            if key in h5file["data"][step]:
                dset = h5file["data"][step][key]
                dset.refresh()
                return np.array(dset)
            return default

        return TDGLData(
            **{field.name: get(field.name) for field in dataclasses.fields(TDGLData)}
        )

    def to_hdf5(self, h5group: h5py.Group) -> None:
        """Save a :class:`TDGLData` instance to an :class:`h5py.Group`.

        Args:
            h5group: An open :class:`h5py.Group` in which to save the data.
        """
        group = h5group.create_group(str(self.step))
        for key, value in dataclasses.asdict(self).items():
            if key in ["step"]:
                continue
            if key in ["state"]:
                group.attrs.update(value)
            else:
                group[key] = value

    def __eq__(self, other: Any) -> bool:
        return dataclass_equals(self, other)


@dataclasses.dataclass(eq=False)
class DynamicsData:
    """A container for the measured dynamics of a TDGL solution,
    measured at each time step in the simulation.

    Args:
        dt: The solver time step, :math:`\\Delta t^{n}`.
        time: The solver time, a derived attribute which is equal to the cumulative
            sum of the time step.
        mu: The electric potential, :math:`\\mu`.
        theta: The phase of the order parameter, :math:`\\theta=\\arg\\psi`
        screening_iterations: The number of screening iterations performed at each
            time step.
    """

    dt: np.ndarray
    time: np.ndarray = dataclasses.field(init=False)
    mu: Union[np.ndarray, None] = None
    theta: Union[np.ndarray, None] = None
    screening_iterations: Union[np.ndarray, None] = None

    def __post_init__(self):
        self.time = np.cumsum(self.dt)

    def time_slice(self, tmin: float = -np.inf, tmax: float = np.inf) -> np.ndarray:
        """Returns the integer indices corresponding to the specified time window.

        Args:
            tmin: The minimum of the time window.
            tmax: The maximum of the time window.

        Returns:
            An array of indices for the time window.
        """
        ts = self.time
        (indices,) = np.where((ts >= tmin) & (ts <= tmax))
        return indices

    def closest_time(self, time: float) -> int:
        """Returns the index of the time step closest to ``time``.

        Args:
            time: The time for which to find the index.

        Returns:
            The index of the time step closest to ``time``.
        """
        return np.argmin(np.abs(self.time - time))

    def voltage(self, i: int = 0, j: int = 1) -> np.ndarray:
        """Returns the voltage, i.e., the electric potential difference
        between probe points ``i`` and ``j``, as a function of time.

        Args:
            i: Index for the first probe point.
            j: Index for the second probe point.

        Returns:
            The voltage :math:`V_{ij}(t)=\\mu_i(t)-\\mu_j(t)`
        """
        if self.mu is None:
            raise ValueError("No voltage data available.")
        if self.mu.shape[0] == 1:
            raise ValueError("The solution has only one probe point.")
        return self.mu[i] - self.mu[j]

    def phase_difference(self, i: int = 0, j: int = 1) -> np.ndarray:
        """Returns the phase difference between probe points ``i`` and ``j``
        as a function of time.

        Args:
            i: Index for the first probe point.
            j: Index for the second probe point.

        Returns:
            The phase difference :math:`\\Delta\\theta_{ij}(t)=\\theta_i(t)-\\theta_j(t)`,
            where :math:`\\theta=\\arg\\psi`.
        """
        if self.theta is None:
            raise ValueError("No phase data available.")
        if self.theta.shape[0] == 1:
            raise ValueError("The solution has only one probe point.")
        return self.theta[i] - self.theta[j]

    def mean_voltage(
        self, i: int = 0, j: int = 1, tmin: float = -np.inf, tmax: float = np.inf
    ) -> float:
        """Returns the time-averaged voltage :math:`\\langle \\Delta\\mu \\rangle`
        over the specified time interval.

        .. math::

            \\langle V_{i,j} \\rangle =
            \\frac{\\sum_n V_{i,j}^{n}\\cdot\\Delta t^{n}}{\\sum_n\\Delta t^{n}}

        Args:
            i: Index for the first probe point.
            j: Index for the second probe point.
            tmin: The minimum of the time window over which to average.
            tmax: The maximum of the time window over which to average.

        Returns:
            The time-averaged voltage over the specified time window.
        """
        if self.mu is None:
            raise ValueError("No voltage data available.")
        indices = self.time_slice(tmin, tmax)
        return np.average(self.voltage(i, j)[indices], weights=self.dt[indices])

    def resample(self, num_points: Union[int, None] = None) -> "DynamicsData":
        """Re-sample the dynamics to a uniform grid using linear interpolation.

        Args:
            num_points: The number of points to interpolate to.

        Returns:
            A new :class:`DynamicsData` instance with the re-sampled data.
        """
        time = self.time
        if num_points is None:
            num_points = len(time)
        ts = np.linspace(time.min(), time.max(), num_points)
        mu = theta = None
        if self.mu is not None:
            mu = np.array([np.interp(ts, time, val) for val in self.mu])
        if self.theta is not None:
            theta = np.array([np.interp(ts, time, val) for val in self.theta])
        return DynamicsData(dt=(ts[1] - ts[0]) * np.ones_like(ts), mu=mu, theta=theta)

    def plot(
        self,
        i: int = 0,
        j: int = 1,
        tmin: float = -np.inf,
        tmax: float = +np.inf,
        grid: bool = True,
        mean_voltage: bool = True,
        labels: bool = True,
        legend: bool = False,
    ) -> Tuple[plt.Figure, Sequence[plt.Axes]]:
        """Plot the voltage and phase difference over the specified time window.

        Args:
            i: Index for the first probe point.
            j: Index for the second probe point.
            tmin: The minimum of the time window to plot.
            tmax: The maximum of the time window to plot.
            grid: Whether to add grid lines to the plots.
            mean_voltage: Whether to plot a horizontal line at the mean voltage.
            labels: Whether to include axis labels.
            legend: Whether to include a legend.

        Returns:
            matplotlib figure and axes.
        """
        fig, axes = plt.subplots(2, 1, sharex=True)
        ax, bx = axes
        ax.grid(grid)
        bx.grid(grid)
        ts = self.time
        vs = self.voltage(i, j)
        phases = np.unwrap(self.phase_difference(i, j)) / np.pi
        indices = self.time_slice(tmin, tmax)
        ax.plot(ts[indices], vs[indices])
        if mean_voltage:
            ax.axhline(
                self.mean_voltage(i=i, j=j, tmin=tmin, tmax=tmax),
                label="Mean voltage",
                color="k",
                ls="--",
            )
        bx.plot(ts[indices], phases[indices])
        if labels:
            ax.set_ylabel(f"Voltage\n$\\Delta\\mu_{{{i},{j}}}$ [$V_0$]")
            bx.set_xlabel("Time, $t$ [$\\tau_0$]")
            bx.set_ylabel(f"Phase difference\n$\\Delta\\theta_{{{i},{j}}}/\\pi$")
        if legend:
            ax.legend(loc=0)
        return fig, axes

    def plot_dt(
        self,
        tmin: float = -np.inf,
        tmax: float = +np.inf,
        grid: bool = True,
        labels: bool = True,
        **histogram_kwargs,
    ) -> Tuple[plt.Figure, Tuple[plt.Axes, plt.Axes]]:
        """Plots the time step :math:`\\Delta t^{n}` vs. time and a histogram of
        :math:`\\Delta t^{n}`.

        Args:
            tmin: The minimum of the time window to plot.
            tmax: The maximum of the time window to plot.
            grid: Whether to add grid lines to the plots.
            labels: Whether to include axis labels.
            histogram_kwargs: Passed to plt.Axes.hist().

        Returns:
            matplotlib figure and two axes.
        """
        fig, (ax, bx) = plt.subplots(1, 2, gridspec_kw=dict(width_ratios=[2, 1]))
        ax.sharey(bx)
        ax.grid(grid)
        bx.grid(grid)
        ts = self.time
        indices = self.time_slice(tmin, tmax)
        ax.plot(ts[indices], self.dt[indices])
        histogram_kwargs = histogram_kwargs.copy()
        histogram_kwargs.setdefault("bins", 101)
        histogram_kwargs.setdefault("density", True)
        histogram_kwargs["orientation"] = "horizontal"
        bx.hist(self.dt[indices], **histogram_kwargs)
        if labels:
            ax.set_xlabel("Time, $t$ [$\\tau_0$]")
            ax.set_ylabel("Time step, $\\Delta t$ [$\\tau_0$]")
            if histogram_kwargs.get("density", False):
                bx.set_xlabel("Density")
            else:
                bx.set_xlabel("Counts per bin")
        fig.tight_layout()
        return fig, (ax, bx)

    @staticmethod
    def from_hdf5(
        h5file: Union[h5py.File, h5py.Group],
        step_min: Union[int, None] = None,
        step_max: Union[int, None] = None,
    ) -> "DynamicsData":
        """Load a :class:`DynamicsData` instance from an output :class:`h5py.File`.

        Args:
            h5file: An open HDF5 output file.
            step_min: The minimum solve step to load.
            step_max: The maximum solve step to load.

        Returns:
            A new :class:`DynamicsData` instance.
        """
        iterations = None
        if "theta" in h5file:
            # Load from DynamicsData.to_hdf5()
            dt = np.array(h5file["dt"])
            mu = screening_iterations = None
            theta = np.array(h5file["theta"])
            if "mu" in h5file:
                mu = np.array(h5file["mu"])
            if "screening_iterations" in h5file:
                iterations = np.array(h5file["screening_iterations"])
        else:
            dts = []
            mus = []
            thetas = []
            screening_iterations = []
            if step_min is None:
                step_min, step_max = get_data_range(h5file)
            for i in range(step_min, step_max + 1):
                grp = h5file[f"data/{i}"]
                if "running_state" not in grp:
                    continue
                grp = grp["running_state"]
                dts.append(np.array(grp["dt"]))
                if "mu" in grp:
                    mus.append(np.array(grp["mu"]))
                if "theta" in grp:
                    thetas.append(np.array(grp["theta"]))
                if "screening_iterations" in grp:
                    screening_iterations.append(np.array(grp["screening_iterations"]))
            dt = np.concatenate(dts)
            mask = dt > 0
            dt = dt[mask]
            mu = theta = iterations = None
            if mus:
                mu = np.concatenate(mus, axis=1)[..., mask]
            if thetas:
                theta = np.concatenate(thetas, axis=1)[..., mask]
            if screening_iterations:
                iterations = np.concatenate(screening_iterations)[mask]
        return DynamicsData(
            dt,
            mu=mu,
            theta=theta,
            screening_iterations=iterations,
        )

    def to_hdf5(self, h5group: h5py.Group) -> None:
        """Save a :class:`DynamicsData` instance to an :class:`h5py.Group`.

        Args:
            h5group: An open :class:`h5py.Group` in which to save the data.
        """
        h5group["dt"] = self.dt
        if self.mu is not None:
            h5group["mu"] = self.mu
        if self.theta is not None:
            h5group["theta"] = self.theta
        if self.screening_iterations is not None:
            h5group["screening_iterations"] = self.screening_iterations

    @staticmethod
    def from_solution(
        solution_path: str,
        probe_points: Optional[Sequence[Tuple[float, float]]] = None,
        progress_bar: bool = False,
    ) -> "DynamicsData":
        """Load :class:`DynamicsData` from the saved time steps of a :class:`tdgl.Solution`.

        Args:
            solution_path: Path to the :class:`tdgl.Solution`
            probe_points: The probe coordinates for which to extract dynamics.
                If ``None``, defaults to ``solution.device.probe_points``.
            progress_bar: Show a progress bar while loading data.

        Returns:
            A new :class:`DynamicsData` instance
        """
        from .solution import Solution

        solution = Solution.from_hdf5(solution_path)
        device = solution.device
        mesh = device.mesh
        if probe_points is None:
            probe_points = device.probe_points
        if probe_points is None:
            raise ValueError("No probe points were provided.")
        probe_points = np.asarray(probe_points).squeeze()
        if probe_points.ndim != 2 or probe_points.shape[1] != 2:
            raise ValueError(
                f"Probe points must have shape (n, 2), got {probe_points.shape}."
            )
        if not device.contains_points(probe_points).all():
            raise ValueError("All probe points must lie within the film.")

        xi = device.coherence_length.magnitude
        probe_point_indices = [mesh.closest_site(xy) for xy in probe_points / xi]
        step_min, step_max = solution.data_range

        num_probes = len(probe_points)
        num_steps = step_max - step_min + 1
        times = np.zeros(num_steps)
        mus = np.zeros((num_probes, num_steps))
        thetas = np.zeros((num_probes, num_steps))

        with h5py.File(solution_path, "r") as h5file:
            for i in tqdm(
                range(step_min, step_max + 1),
                desc="Time steps",
                disable=(not progress_bar),
            ):
                grp = h5file[f"data/{i}"]
                times[i] = float(grp.attrs["time"])
                mus[:, i] = np.array(grp["mu"])[probe_point_indices]
                thetas[:, i] = np.angle(np.array(grp["psi"]))[probe_point_indices]

        return DynamicsData(dt=np.diff(times), mu=mus, theta=thetas)

    def __eq__(self, other: Any) -> bool:
        return dataclass_equals(self, other)


def get_current_through_paths(
    solution_path: str,
    paths: Union[np.ndarray, List[np.ndarray]],
    dataset: Optional[str] = None,
    interp_method: Literal["linear", "cubic"] = "linear",
    units: Optional[str] = None,
    with_units: bool = True,
    progress_bar: bool = True,
) -> Tuple[np.ndarray, Union[np.ndarray, List[np.ndarray]]]:
    """Calculates the current through one or more paths for each saved time step.

    Args:
        solution_path: Path to the solution HDF5 file.
        paths: A list of ``(n, 2)`` arrays of ``(x, y)`` coordinates defining
            the paths. A single ``(n, 2)`` array is also allowed.
        dataset: ``None``, ``"supercurrent"``, or ``"normal_current"``.
            ``None`` indicates the total current.
        interp_method: Interpolation method: either "linear" or "cubic".
        units: The current units to return.
        with_units: Whether to return a :class:`pint.Quantity` with units attached.
        progress_bar: Whether to display a progress bar.

    Returns:
        ``(times, currents)``, where ``currents`` is a list of arrays of the
        time-dependent current through each path. If ``paths`` is given as a
        single array, ``currents`` will be returned as a single array.
    """
    from .solution import Solution

    solution = Solution.from_hdf5(solution_path)
    device = solution.device
    tri = device.triangulation
    ureg = device.ureg

    valid_methods = ("linear", "cubic")
    if interp_method not in valid_methods:
        raise ValueError(
            f"Interpolation method must be one of {valid_methods} (got {interp_method})."
        )
    interp_type = {
        "linear": mtri.LinearTriInterpolator,
        "cubic": mtri.CubicTriInterpolator,
    }[interp_method]

    valid_datasets = ("supercurrent", "normal_current", None)
    if dataset not in valid_datasets:
        raise ValueError(
            f"Dataset name must be one of {valid_datasets} (got {dataset})."
        )

    if units is None:
        units = solution.current_units

    length_units = ureg(device.length_units)

    if isinstance(paths, np.ndarray):
        paths = [paths]
    paths = [np.asarray(p) for p in paths]
    edge_positions = []
    edge_lengths = []
    unit_normals = []
    in_device = []
    for path in paths:
        edge_positions.append((path[:-1] + path[1:]) / 2)
        lengths, normals = path_vectors(path)
        edge_lengths.append(lengths)
        unit_normals.append(normals)
        in_device.append(device.contains_points(edge_positions[-1]))

    step_min, step_max = solution.data_range
    times = solution.times
    raw_currents = [np.zeros_like(times) for _ in paths]
    with h5py.File(solution_path, "r") as h5file:
        for i in tqdm(
            range(step_min, step_max + 1), desc="Time steps", disable=(not progress_bar)
        ):
            grp = h5file[f"data/{i}"]
            if dataset is None:
                K = np.array(grp["normal_current"]) + np.array(grp["supercurrent"])
            else:
                K = np.array(grp[dataset])
            K = device.mesh.get_quantity_on_site(K)
            Kx_interp = interp_type(tri, K[:, 0])
            Ky_interp = interp_type(tri, K[:, 1])
            for j, (path, lengths, normals, ix) in enumerate(
                zip(paths, edge_lengths, unit_normals, in_device)
            ):
                Kx_path = Kx_interp(path[:, 0], path[:, 1]).data
                Ky_path = Ky_interp(path[:, 0], path[:, 1]).data
                K_path = np.array([Kx_path, Ky_path]).T
                K_path[~np.isfinite(K_path).all(axis=1)] = 0
                # Evaluate the sheet current at the edge centers
                K_edge = (K_path[:-1] + K_path[1:]) / 2
                K_dot_n = (K_edge * normals).sum(axis=1)
                raw_currents[j][i] = np.trapz((K_dot_n * lengths)[ix])

    currents = []
    for current in raw_currents:
        J = current * (device.K0 * length_units).to(units)
        if not with_units:
            J = J.magnitude
        currents.append(J)

    if len(currents) == 1:
        currents = currents[0]

    return times, currents



import numba
import numpy as np

try:
    import cupy  # type: ignore
    import cupyx  # type: ignore
except ImportError:
    cupy = None
    cupyx = None


@numba.njit(fastmath=True, parallel=True)
def get_A_induced_numba(
    J_site: np.ndarray,
    site_areas: np.ndarray,
    sites: np.ndarray,
    edge_centers: np.ndarray,
    A_induced: np.ndarray,
) -> None:
    """Calculates the induced vector potential on the mesh edges.

    Args:
        J_site: The current density on the sites, shape ``(n, )``
        site_areas: The mesh site areas, shape ``(n, )``
        sites: The mesh site coordinates, shape ``(n, 2)``
        edge_centers: The coordinates of the edge centers, shape ``(m, 2)``
        A_induced: The output array, shape ``(m, 2)``
    """
    assert J_site.ndim == 2
    assert J_site.shape[1] == 2
    assert sites.shape == J_site.shape
    assert edge_centers.ndim == 2
    assert edge_centers.shape[1] == 2
    for i in numba.prange(edge_centers.shape[0]):
        for k in range(J_site.shape[1]):
            tmp = 0.0
            for j in range(J_site.shape[0]):
                dx = edge_centers[i, 0] - sites[j, 0]
                dy = edge_centers[i, 1] - sites[j, 1]
                dr = np.sqrt(dx * dx + dy * dy)
                tmp += J_site[j, k] * site_areas[j] / dr
            A_induced[i, k] = tmp


get_A_induced_cupy = None

if cupy is not None:

    @cupyx.jit.rawkernel()
    def get_A_induced_cupy(
        J_site: cupy.ndarray,
        site_areas: cupy.ndarray,
        sites: cupy.ndarray,
        edge_centers: cupy.ndarray,
        A_induced: cupy.ndarray,
    ) -> None:
        """Calculates the induced vector potential on the mesh edges.

        Args:
            J_site: The current density on the sites, shape ``(n, )``
            site_areas: The mesh site areas, shape ``(n, )``
            sites: The mesh site coordinates, shape ``(n, 2)``
            edge_centers: The coordinates of the edge centers, shape ``(m, 2)``
            A_induced: The induced vector potential on the mesh edges,
                i.e. the output array, shape ``(m, 2)``.
        """
        i, k = cupyx.jit.grid(2)
        if i < edge_centers.shape[0] and k < J_site.shape[1]:
            tmp = 0.0
            for j in cupyx.jit.range(sites.shape[0]):
                dx = edge_centers[i, 0] - sites[j, 0]
                dy = edge_centers[i, 1] - sites[j, 1]
                dr = cupy.sqrt(dx * dx + dy * dy)
                tmp += J_site[j, k] * site_areas[j] / dr
            A_induced[i, k] = tmp



import numba
import numpy as np


@numba.njit(fastmath=True, parallel=True)
def sqeuclidean_distance_2d(XA: np.ndarray, XB: np.ndarray):
    """Squared Euclidean pointwise distance between two 2D arrays."""
    out = np.empty((XA.shape[0], XB.shape[0]), dtype=XA.dtype)
    for i in numba.prange(XA.shape[0]):
        for j in range(XB.shape[0]):
            dx = XA[i, 0] - XB[j, 0]
            dy = XA[i, 1] - XB[j, 1]
            out[i, j] = dx * dx + dy * dy
    return out


@numba.njit(fastmath=True, parallel=True)
def sqeuclidean_distance_3d(XA: np.ndarray, XB: np.ndarray):
    """Squared Euclidean pointwise distance between two 3D arrays."""
    out = np.empty((XA.shape[0], XB.shape[0]), dtype=XA.dtype)
    for i in numba.prange(XA.shape[0]):
        for j in range(XB.shape[0]):
            dx = XA[i, 0] - XB[j, 0]
            dy = XA[i, 1] - XB[j, 1]
            dz = XA[i, 2] - XB[j, 2]
            out[i, j] = dx * dx + dy * dy + dz * dz
    return out


@numba.njit(fastmath=True, parallel=True)
def euclidean_distance_2d(XA: np.ndarray, XB: np.ndarray):
    """Euclidean pointwise distance between two 2D arrays."""
    out = np.empty((XA.shape[0], XB.shape[0]), dtype=XA.dtype)
    for i in numba.prange(XA.shape[0]):
        for j in range(XB.shape[0]):
            dx = XA[i, 0] - XB[j, 0]
            dy = XA[i, 1] - XB[j, 1]
            out[i, j] = np.sqrt(dx * dx + dy * dy)
    return out


@numba.njit(fastmath=True, parallel=True)
def euclidean_distance_3d(XA: np.ndarray, XB: np.ndarray):
    """Euclidean pointwise distance between two 3D arrays."""
    out = np.empty((XA.shape[0], XB.shape[0]), dtype=XA.dtype)
    for i in numba.prange(XA.shape[0]):
        for j in range(XB.shape[0]):
            dx = XA[i, 0] - XB[j, 0]
            dy = XA[i, 1] - XB[j, 1]
            dz = XA[i, 2] - XB[j, 2]
            out[i, j] = np.sqrt(dx * dx + dy * dy + dz * dz)
    return out


def cdist(XA: np.ndarray, XB: np.ndarray, metric: str = "euclidean"):
    """Pointwise distance between observations in 2D or 3D space.

    This function provides a subset of the functionality of
    ``scipy.spatial.distance.cdist``. The output array has the same dtype
    as the first input, XA.

    Args:
        XA: An (mA, n) array of observations, where n is 2 or 3.
        XB: An (mB, n) array of observations, where n is 2 or 3.
        metric: Either 'euclidean' or 'sqeuclidean'.

    Returns:
        An (mA, mB) matrix of pointwise distances, where ``out[i, j] = dist(XA[i], XB[j])``.
    """
    metrics = ("euclidean", "sqeuclidean")
    if metric not in metrics:
        raise ValueError(f"Metric must be one of {metrics!r}, got {metric!r}.")
    if XA.shape[1] != XB.shape[1]:
        raise ValueError(
            f"XA.shape[1] ({XA.shape[1]}) must be equal to XB.shape[1] ({XB.shape[1]})."
        )
    if XA.shape[1] == 2:
        if metric == "euclidean":
            return euclidean_distance_2d(XA, XB)
        return sqeuclidean_distance_2d(XA, XB)
    elif XA.shape[1] == 3:
        if metric == "euclidean":
            return euclidean_distance_3d(XA, XB)
        return sqeuclidean_distance_3d(XA, XB)
    raise ValueError(f"Excpected shape (n, 2) arrays, got {XA.shape} and {XB.shape}.")



__version_info__ = (0, 8, 0)
__version__ = ".".join(map(str, __version_info__))


def git_version():
    import os.path
    import subprocess

    git_hash = git_date = None

    dirname = os.path.dirname(__file__)

    if not os.path.exists(os.path.join(dirname, os.pardir, ".git")):
        return git_hash, git_date

    try:
        p = subprocess.Popen(
            ["git", "log", "-1", '--format="%H %aI"'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=dirname,
        )
    except FileNotFoundError:
        pass
    else:
        out, err = p.communicate()
        if p.returncode == 0:
            git_hash, git_date = (
                out.decode("utf-8").strip().replace('"', "").split("T")[0].split()
            )

    return git_hash, git_date


git_hash, git_date = git_version()

__git_revision__ = None
if git_hash is not None:
    __git_revision__ = f"{git_hash[:7]} [{git_date}]"



from typing import Callable, Dict, Optional, Union

from ..device.device import Device
from ..solution.solution import Solution
from .options import SolverOptions
from .solver import TDGLSolver


def solve(
    device: Device,
    options: SolverOptions,
    applied_vector_potential: Union[Callable, float] = 0,
    terminal_currents: Union[Callable, Dict[str, float], None] = None,
    disorder_epsilon: Union[float, Callable] = 1,
    seed_solution: Optional[Solution] = None,
) -> Union[Solution, None]:
    """Solve a TDGL model.

    Args:
        device: The :class:`tdgl.Device` to solve.
        options: An instance :class:`tdgl.SolverOptions` specifying the solver
            parameters.
        applied_vector_potential: A function or :class:`tdgl.Parameter` that computes
            the applied vector potential as a function of position ``(x, y, z)``,
            or of position and time ``(x, y, z, *, t)``. If a float ``B`` is given,
            the applied vector potential will be that of a uniform magnetic field with
            strength ``B`` ``field_units``.
        terminal_currents: A dict of ``{terminal_name: current}`` or a callable with signature
            ``func(time: float) -> {terminal_name: current}``, where ``current`` is a float
            in units of ``current_units`` and ``time`` is the dimensionless time.
        disorder_epsilon: A float in range [-1, 1], or a callable with signature
            ``disorder_epsilon(r: Tuple[float, float]) -> epsilon``, where ``epsilon``
            is a float in range [-1, 1]. Setting
            :math:`\\epsilon(\\mathbf{r})=T_c(\\mathbf{r})/T_c - 1 < 1` suppresses the
            critical temperature at position :math:`\\mathbf{r}`, which can be used
            to model inhomogeneity.
        seed_solution: A :class:`tdgl.Solution` instance to use as the initial state
            for the simulation.

    Returns:
        A :class:`tdgl.Solution` instance. Returns ``None`` if the simulation was
        cancelled during the thermalization stage.
    """
    solver = TDGLSolver(
        device=device,
        options=options,
        applied_vector_potential=applied_vector_potential,
        terminal_currents=terminal_currents,
        disorder_epsilon=disorder_epsilon,
        seed_solution=seed_solution,
    )
    return solver.solve()



import numba
import numpy as np
import pytest

try:
    import cupy  # type: ignore
except ImportError:
    cupy = None

import tdgl
from tdgl.geometry import box, circle
from tdgl.solver.options import SolverOptionsError


@pytest.mark.parametrize("current", [5.0, lambda t: 10])
@pytest.mark.parametrize("field", [0, 1])
@pytest.mark.parametrize(
    "terminal_psi, time_dependent, gpu, vectorized",
    [(0, True, False, True), (1, False, False, False), (1, True, True, True)],
)
def test_source_drain_current(
    transport_device,
    current,
    field,
    terminal_psi,
    time_dependent,
    gpu,
    vectorized,
):
    device = transport_device
    total_time = 10
    skip_time = 1

    if gpu and cupy is None:
        options = tdgl.SolverOptions(
            solve_time=total_time,
            skip_time=skip_time,
            field_units="uT",
            current_units="uA",
            save_every=100,
            terminal_psi=terminal_psi,
            gpu=gpu,
        )
        with pytest.raises(SolverOptionsError):
            options.validate()
        return

    options = tdgl.SolverOptions(
        solve_time=total_time,
        skip_time=skip_time,
        field_units="uT",
        current_units="uA",
        save_every=100,
        terminal_psi=terminal_psi,
        gpu=gpu,
    )

    options.sparse_solver = "unknown"
    with pytest.raises(SolverOptionsError):
        options.validate()
    options.sparse_solver = "superlu"
    options.validate()

    if callable(current):

        def terminal_currents(t):
            return dict(source=current(0), drain=-current(0))

    else:
        terminal_currents = dict(source=current, drain=-current)

    with pytest.raises(ValueError):
        solution = tdgl.solve(
            device,
            options,
            disorder_epsilon=2,
            applied_vector_potential=field,
            terminal_currents=terminal_currents,
        )

    if vectorized:

        def disorder_epsilon(r):
            return 1.0 * np.ones(len(r))

    else:

        def disorder_epsilon(r):
            return 1.0

    if time_dependent:
        ramp = tdgl.sources.LinearRamp(tmin=1, tmax=8)
        constant_field = tdgl.sources.ConstantField(
            field,
            field_units=options.field_units,
            length_units=device.length_units,
        )
        field = ramp * constant_field
        field = constant_field * ramp

        _disorder_epsilon = disorder_epsilon

        def disorder_epsilon(r, *, t, vectorized=vectorized):
            return _disorder_epsilon(r)

    solution = tdgl.solve(
        device,
        options,
        disorder_epsilon=disorder_epsilon,
        applied_vector_potential=field,
        terminal_currents=terminal_currents,
    )

    if callable(current):
        current = current(0)

    ys = np.linspace(-5, 5, 501)
    measured_currents = []
    for x0 in [-8, -2, 0, 2, 8]:
        coords = np.array([x0 * np.ones_like(ys), ys]).T
        measured_currents.append(
            solution.current_through_path(coords, with_units=False)
        )
    measured_currents = np.array(measured_currents)
    assert np.allclose(measured_currents, current, rtol=0.1)


@pytest.fixture
def screening_device() -> tdgl.Device:
    length_units = "um"
    xi = 0.1
    london_lambda = 0.075
    thickness = 0.05

    height = 1
    width = 2

    layer = tdgl.Layer(
        coherence_length=xi, london_lambda=london_lambda, thickness=thickness
    )
    film = tdgl.Polygon("film", points=tdgl.geometry.box(width, height, points=301))
    device = tdgl.Device(
        "bar",
        layer=layer,
        film=film,
        length_units=length_units,
    )
    device.make_mesh(max_edge_length=xi / 2, smooth=100)
    return device


def test_screening(screening_device: tdgl.Device):
    numba.set_num_threads(2)
    device = screening_device

    fluxoid_curves = [
        circle(0.25, center=(0, 0)),
        circle(0.1, center=(0.15, 0.25)),
        circle(0.3, center=(0.6, -0.1)),
        box(0.5, center=(-0.5, 0)),
        box(0.5, center=(-0.6, -0.2)),
    ]

    options = tdgl.SolverOptions(
        solve_time=2,
        field_units="mT",
        current_units="uA",
        include_screening=False,
        monitor=True,
    )

    no_screening_solution = tdgl.solve(device, options, applied_vector_potential=0.1)
    K = no_screening_solution.current_density
    K_max = np.sqrt(K[:, 0] ** 2 + K[:, 1] ** 2).max().to("uA / um").magnitude

    assert np.isclose(K_max, 450, rtol=5e-2)

    for curve in fluxoid_curves:
        fluxoid = no_screening_solution.polygon_fluxoid(curve)
        total_fluxoid = sum(fluxoid).magnitude
        error = abs(total_fluxoid / fluxoid.flux_part.magnitude)
        assert error > 1

    options.include_screening = True
    options.screening_tolerance = 1e-6
    options.dt_max = 1e-3

    screening_solution = tdgl.solve(device, options, applied_vector_potential=0.1)
    K = screening_solution.current_density
    K_max = np.sqrt(K[:, 0] ** 2 + K[:, 1] ** 2).max().to("uA / um").magnitude
    assert np.isclose(K_max, 270, rtol=2e-2)

    for curve in fluxoid_curves:
        fluxoid = screening_solution.polygon_fluxoid(curve)
        total_fluxoid = sum(fluxoid).magnitude
        error = abs(total_fluxoid / fluxoid.flux_part.magnitude)
        assert error < 5e-2



import os
import sys
from typing import Any, Dict, Literal, Optional, Sequence, Tuple, Union

import h5py
import matplotlib.pyplot as plt
import numpy as np

from ..device.device import Device
from .common import DEFAULT_QUANTITIES, PLOT_DEFAULTS, Quantity, auto_grid
from .io import get_plot_data


def monitor_solution(
    h5path: str,
    quantities: Union[str, Sequence[str]] = DEFAULT_QUANTITIES,
    update_interval: float = 1.0,
    shading: Literal["flat", "gouraud"] = "gouraud",
    max_cols: int = 4,
    dimensionless: bool = False,
    xlim: Optional[Tuple[float, float]] = None,
    ylim: Optional[Tuple[float, float]] = None,
    autoscale: bool = True,
    figure_kwargs: Optional[Dict[str, Any]] = None,
) -> None:
    """Plots the results of a simulation while it is running.

    Args:
        h5path: Path to the temporary HDF5 file generated when running
            the solver with options.monitor=True
        quantities: The names of the quantities to animate
        update_interval: The update interval in seconds
        shading: Shading method, "flat" or "gouraud". See matplotlib.pyplot.tripcolor.
        dimensionless: Use dimensionless units for axes
        xlim: x-axis limits
        ylim: y-axis limits
        autoscale: Autoscale colorbar limits at each frame
    """

    # Try to use the appropriate GUI backend
    plt.switch_backend("agg")
    for candidate in ["macosx", "qt5agg", "qt4agg", "gtk3agg", "tkagg", "wxagg"]:
        try:
            plt.switch_backend(candidate)
            break
        except ImportError:
            continue

    if quantities is None:
        quantities = Quantity.get_keys()
    if isinstance(quantities, str):
        quantities = [quantities]
    quantities = [Quantity.from_key(name.upper()) for name in quantities]
    num_plots = len(quantities)
    figure_kwargs = figure_kwargs or dict()
    figure_kwargs.setdefault("constrained_layout", True)
    default_figsize = (
        3.5 * min(max_cols, num_plots),
        3.0 * max(1, num_plots // max_cols),
    )
    figure_kwargs.setdefault("figsize", default_figsize)
    figure_kwargs.setdefault("sharex", True)
    figure_kwargs.setdefault("sharey", True)

    with h5py.File(h5path, "r", swmr=True, libver="latest") as h5file:
        device = Device.from_hdf5(h5file["solution/device"])
        mesh = device.mesh
        if dimensionless:
            scale = 1
            units_str = "\\xi"
        else:
            scale = device.layer.coherence_length
            units_str = f"{device.ureg(device.length_units).units:~L}"
        x, y = scale * mesh.sites.T

        # Temp data to use in plots
        temp_value = np.ones(len(mesh.sites), dtype=float)
        temp_value[0] = 0
        temp_value[1] = 0.5

        plt.ion()
        fig, axes = auto_grid(num_plots, max_cols=max_cols, **figure_kwargs)
        fig.canvas.mpl_connect("close_event", lambda event: sys.exit(0))

        collections = []
        for quantity, ax in zip(quantities, axes.flat):
            ax: plt.Axes
            opts = PLOT_DEFAULTS[quantity]
            collection = ax.tripcolor(
                x,
                y,
                temp_value,
                triangles=mesh.elements,
                shading=shading,
                cmap=opts.cmap,
                vmin=opts.vmin,
                vmax=opts.vmax,
            )
            cbar = fig.colorbar(collection, ax=ax)
            cbar.set_label(opts.clabel)
            ax.set_aspect("equal")
            ax.set_title(quantity.value)
            ax.set_xlabel(f"$x$ [${units_str}$]")
            ax.set_ylabel(f"$y$ [${units_str}$]")
            collections.append(collection)
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)

        fig.suptitle("Step: 0")
        fig.canvas.draw_idle()
        fig.canvas.start_event_loop(1e-3)

        vmins = [+np.inf for _ in quantities]
        vmaxs = [-np.inf for _ in quantities]

        prev_step = 0

        def update():
            grp = h5file["data/-1"]
            step = np.array(grp["step"])[0]
            nonlocal prev_step
            if step == prev_step:
                fig.canvas.start_event_loop(update_interval)
                return
            prev_step = step
            time = np.array(grp["time"])[0]
            dt = np.array(grp["dt"])[0]
            fig.suptitle(f"Step: {step}, time: {time:.2e}, dt: {dt:.2e}")
            for i, (quantity, collection) in enumerate(zip(quantities, collections)):
                opts = PLOT_DEFAULTS[quantity]
                values, _, _ = get_plot_data(h5file, mesh, quantity, -1)
                mask = np.abs(values - np.mean(values)) <= 6 * np.std(values)
                if opts.vmin is None:
                    if autoscale:
                        vmins[i] = np.min(values[mask])
                    else:
                        vmins[i] = min(vmins[i], np.min(values[mask]))
                else:
                    vmins[i] = opts.vmin
                if opts.vmax is None:
                    if autoscale:
                        vmaxs[i] = np.max(values[mask])
                    else:
                        vmaxs[i] = max(vmaxs[i], np.max(values[mask]))
                else:
                    vmaxs[i] = opts.vmax
                if opts.symmetric:
                    vmax = max(abs(vmins[i]), abs(vmaxs[i]))
                    vmaxs[i] = vmax
                    vmins[i] = -vmax
                if shading == "flat":
                    # https://stackoverflow.com/questions/40492511/set-array-in-tripcolor-bug
                    values = values[mesh.elements].mean(axis=1)
                collection.set_array(values)
                collection.set_clim(vmins[i], vmaxs[i])
            fig.canvas.draw_idle()
            fig.canvas.start_event_loop(update_interval)

        try:
            while True:
                if os.path.exists(h5path):
                    update()
                else:
                    plt.close(fig)
        except KeyboardInterrupt:
            plt.close(fig)



import warnings
from typing import Callable, Tuple, Union

import numpy as np
import scipy.sparse as sp

try:
    import cupy  # type: ignore
    from cupyx.scipy.sparse import csc_matrix, csr_matrix  # type: ignore
    from cupyx.scipy.sparse.linalg import factorized  # type: ignore
except ImportError:
    cupy = None

from ..solver.options import SparseSolver
from .mesh import Mesh


def _get_spmatrix_offsets_cupy(spmatrix, i, j):
    """Calculates the sparse matrix offsets for a set of rows ``i`` and columns ``j``."""
    # See _set_many() at
    # https://github.com/cupy/cupy/blob/5c32e40af32f6f9627e09d47ecfeb7e9281ccab2/cupyx/scipy/sparse/_compressed.py#L525
    i, j, M, N = spmatrix._prepare_indices(i, j)
    new_sp = csr_matrix(
        (
            cupy.arange(spmatrix.nnz, dtype=cupy.float32),
            spmatrix.indices,
            spmatrix.indptr,
        ),
        shape=(M, N),
    )
    offsets = new_sp._get_arrayXarray(i, j, not_found_val=-1).astype(cupy.int32).ravel()
    return offsets, (i, j, M, N)


def _spmatrix_set_many(spmatrix, i, j, x):
    """spmatrix.__setitem__()"""
    if sp.issparse(spmatrix):
        spmatrix[i, j] = x
        return

    i, j = spmatrix._swap(i, j)
    offsets, (i, j, M, N) = _get_spmatrix_offsets_cupy(spmatrix, i, j)

    mask = offsets > -1
    # update where possible
    spmatrix.data[offsets[mask]] = x[mask]

    if not mask.all():
        # only insertions remain
        mask = ~mask
        i = i[mask]
        i[i < 0] += M
        j = j[mask]
        j[j < 0] += N
        spmatrix._insert_many(i, j, x[mask])


def build_divergence(mesh: Mesh) -> sp.csr_array:
    """Build the divergence matrix that takes the divergence of a function living
    on the edges onto the sites.

    Args:
        mesh: The mesh.

    Returns:
        The divergence matrix.
    """
    edge_mesh = mesh.edge_mesh
    # Indices for each edge
    edge_indices = np.arange(len(edge_mesh.edges))
    # Compute the weights for each edge
    weights = edge_mesh.dual_edge_lengths
    # Rows and cols to update
    edges0 = edge_mesh.edges[:, 0]
    edges1 = edge_mesh.edges[:, 1]
    rows = np.concatenate([edges0, edges1])
    cols = np.concatenate([edge_indices, edge_indices])
    values = np.concatenate(
        [weights / mesh.areas[edges0], -weights / mesh.areas[edges1]]
    )
    return sp.csr_array(
        (values, (rows, cols)), shape=(len(mesh.sites), len(edge_mesh.edges))
    )


def build_gradient(
    mesh: Mesh,
    link_exponents: Union[np.ndarray, None] = None,
    weights: Union[np.ndarray, None] = None,
) -> sp.csr_array:
    """Build the gradient for a function living on the sites onto the edges.

    Args:
        mesh: The mesh.
        link_exponents: The value is integrated, exponentiated and used as
            a link variable.

    Returns:
        The gradient matrix.
    """
    edge_mesh = mesh.edge_mesh
    edge_indices = np.arange(len(edge_mesh.edges))
    if weights is None:
        weights = 1 / edge_mesh.edge_lengths
    if link_exponents is None:
        link_variable_weights = np.ones(len(weights))
    else:
        link_variable_weights = np.exp(
            -1j * np.einsum("ij, ij -> i", link_exponents, edge_mesh.directions)
        )
    rows = np.concatenate([edge_indices, edge_indices])
    cols = np.concatenate([edge_mesh.edges[:, 1], edge_mesh.edges[:, 0]])
    values = np.concatenate([link_variable_weights * weights, -weights])
    return sp.csr_array(
        (values, (rows, cols)), shape=(len(edge_mesh.edges), len(mesh.sites))
    )


def build_laplacian(
    mesh: Mesh,
    link_exponents: Union[np.ndarray, None] = None,
    fixed_sites: Union[np.ndarray, None] = None,
    free_rows: Union[np.ndarray, None] = None,
    fixed_sites_eigenvalues: float = 1,
    weights: Union[np.ndarray, None] = None,
) -> Tuple[sp.csc_array, np.ndarray]:
    """Build a Laplacian matrix on a given mesh.

    The default boundary condition is homogenous Neumann conditions. To get
    Dirichlet conditions, add fixed sites. To get non-homogenous Neumann condition,
    the flux needs to be specified using a Neumann boundary Laplacian matrix.

    Args:
        mesh: The mesh.
        link_exponents: The value is integrated, exponentiated and used as a
            link variable.
        fixed_sites: The sites to hold fixed.
        fixed_sites_eigenvalues: The eigenvalues for the fixed sites.

    Returns:
        The Laplacian matrix and indices of non-fixed rows.
    """
    if fixed_sites is None:
        fixed_sites = np.array([], dtype=int)

    edge_mesh = mesh.edge_mesh
    if weights is None:
        weights = edge_mesh.dual_edge_lengths / edge_mesh.edge_lengths
    if link_exponents is None:
        link_variable_weights = np.ones(len(weights))
    else:
        link_variable_weights = np.exp(
            -1j * np.einsum("ij, ij -> i", link_exponents, edge_mesh.directions)
        )
    edges0 = edge_mesh.edges[:, 0]
    edges1 = edge_mesh.edges[:, 1]
    rows = np.concatenate([edges0, edges1, edges0, edges1])
    cols = np.concatenate([edges1, edges0, edges0, edges1])
    areas0 = mesh.areas[edges0]
    areas1 = mesh.areas[edges1]
    values = np.concatenate(
        [
            weights * link_variable_weights / areas0,
            weights * link_variable_weights.conjugate() / areas1,
            -weights / areas0,
            -weights / areas1,
        ]
    )
    # Exclude all edges connected to fixed sites and set the
    # fixed site diagonal elements separately.
    if free_rows is None:
        free_rows = np.isin(rows, fixed_sites, invert=True)
    rows = rows[free_rows]
    cols = cols[free_rows]
    values = values[free_rows]
    rows = np.concatenate([rows, fixed_sites])
    cols = np.concatenate([cols, fixed_sites])
    values = np.concatenate(
        [values, fixed_sites_eigenvalues * np.ones(len(fixed_sites))]
    )
    laplacian = sp.csc_array(
        (values, (rows, cols)), shape=(len(mesh.sites), len(mesh.sites))
    )
    return laplacian, free_rows


def build_neumann_boundary_laplacian(
    mesh: Mesh, fixed_sites: Union[np.ndarray, None] = None
) -> sp.csr_array:
    """Build extra matrix for the Laplacian to set non-homogenous Neumann
    boundary conditions.

    Args:
        mesh: The mesh.
        fixed_sites: The fixed sites.

    Returns:
        The Neumann boundary matrix.
    """

    edge_mesh = mesh.edge_mesh
    boundary_index = np.arange(len(edge_mesh.boundary_edge_indices))
    # Get the boundary edges which are stored in the beginning of
    # the edge vector
    boundary_edges = edge_mesh.edges[edge_mesh.boundary_edge_indices]
    boundary_edges_length = edge_mesh.edge_lengths[edge_mesh.boundary_edge_indices]
    # Rows and cols to update
    rows = np.concatenate([boundary_edges[:, 0], boundary_edges[:, 1]])
    cols = np.concatenate([boundary_index, boundary_index])
    # The values
    values = np.concatenate(
        [
            boundary_edges_length / (2 * mesh.areas[boundary_edges[:, 0]]),
            boundary_edges_length / (2 * mesh.areas[boundary_edges[:, 1]]),
        ]
    )
    # Build the matrix
    neumann_laplacian = sp.csr_array(
        (values, (rows, cols)), shape=(len(mesh.sites), len(boundary_index))
    )
    # Change the rows corresponding to fixed sites to identity
    if fixed_sites is not None:
        # Convert laplacian to list of lists
        # This makes it quick to do slices
        neumann_laplacian = neumann_laplacian.tolil()
        # Change the rows corresponding to the fixed sites
        neumann_laplacian[fixed_sites, :] = 0

    return neumann_laplacian.tocsr(copy=False)


class MeshOperators:
    """A container for the finite volume operators for a given mesh.

    Args:
        mesh: The :class:`tdgl.finite_volume.Mesh` instance for which to construct
            operators.
        sparse_solver: The sparse solver for which to build mesh operators.
        use_cupy: Use CuPy for linear algebra.
        fixed_sites: The indices of any sites for which the value of :math:`\\psi`
            and :math:`\\mu` are fixed as boundary conditions.
    """

    def __init__(
        self,
        mesh: Mesh,
        sparse_solver: SparseSolver,
        use_cupy: bool = False,
        fixed_sites: Union[np.ndarray, None] = None,
        fix_psi: bool = True,
    ):
        self.mesh = mesh
        self.areas = mesh.areas
        edge_mesh = mesh.edge_mesh
        self.edges = edge_mesh.edges
        self.edge_directions = edge_mesh.directions
        self.use_cupy = use_cupy
        self.sparse_solver = sparse_solver
        self.fixed_sites = fixed_sites
        self.fix_psi = fix_psi
        self.laplacian_free_rows: Union[np.ndarray, None] = None
        self.divergence: Union[sp.spmatrix, None] = None
        self.mu_laplacian: Union[sp.spmatrix, None] = None
        self.mu_boundary_laplacian: Union[sp.spmatrix, None] = None
        self.mu_laplacian_lu: Union[Callable, None] = None
        self.psi_gradient: Union[sp.spmatrix, None] = None
        self.psi_laplacian: Union[sp.spmatrix, None] = None
        self.link_exponents: Union[np.ndarray, None] = None
        # Compute these quantities just once, as they never change.
        self.gradient_weights = 1 / edge_mesh.edge_lengths
        self.laplacian_weights = edge_mesh.dual_edge_lengths / edge_mesh.edge_lengths
        self.gradient_link_rows = np.arange(len(edge_mesh.edges), dtype=int)
        self.gradient_link_cols = edge_mesh.edges[:, 1]
        self.laplacian_link_rows = np.concatenate(
            [edge_mesh.edges[:, 0], edge_mesh.edges[:, 1]]
        )
        self.laplacian_link_cols = np.concatenate(
            [edge_mesh.edges[:, 1], edge_mesh.edges[:, 0]]
        )

    def build_operators(self) -> None:
        """Construct the vector potential-independent operators."""
        mesh = self.mesh
        self.mu_laplacian, _ = build_laplacian(mesh, weights=self.laplacian_weights)
        self.mu_boundary_laplacian = build_neumann_boundary_laplacian(mesh)
        self.mu_gradient = build_gradient(mesh, weights=self.gradient_weights)
        self.divergence = build_divergence(mesh)
        if self.use_cupy:
            assert cupy is not None
            self.mu_boundary_laplacian = csr_matrix(self.mu_boundary_laplacian)
            self.mu_gradient = csr_matrix(self.mu_gradient)
            self.divergence = csr_matrix(self.divergence)
            self.areas = cupy.array(self.areas)
            self.edge_directions = cupy.array(self.edge_directions)
        if self.sparse_solver is SparseSolver.CUPY:
            assert cupy is not None
            self.mu_laplacian = csc_matrix(self.mu_laplacian)
            self.mu_laplacian_lu = factorized(self.mu_laplacian)
        elif self.sparse_solver is SparseSolver.PARDISO:
            self.mu_laplacian_lu = None
        else:
            use_umfpack = self.sparse_solver is SparseSolver.UMFPACK
            sp.linalg.use_solver(useUmfpack=use_umfpack)
            self.mu_laplacian_lu = sp.linalg.factorized(self.mu_laplacian)

    def set_link_exponents(self, link_exponents: np.ndarray) -> None:
        """Set the link variables and construct the covarient gradient
        and Laplacian for psi.

        Args:
            link_exponents: The value is integrated, exponentiated and used as
                a link variable.
        """
        mesh = self.mesh
        xp = cupy if self.use_cupy else np
        self.link_exponents = xp.asarray(link_exponents)
        if self.psi_gradient is None:
            # Build the matrices from scratch
            self.psi_gradient = build_gradient(
                mesh,
                link_exponents=link_exponents,
                weights=self.gradient_weights,
            )
            if self.fix_psi:
                fixed_sites = self.fixed_sites
                free_rows = self.laplacian_free_rows
            else:
                fixed_sites = free_rows = None
            self.psi_laplacian, self.laplacian_free_rows = build_laplacian(
                mesh,
                link_exponents=link_exponents,
                fixed_sites=fixed_sites,
                free_rows=free_rows,
                weights=self.laplacian_weights,
            )
            if self.use_cupy:
                self.psi_gradient = csr_matrix(self.psi_gradient)
                self.psi_laplacian = csr_matrix(self.psi_laplacian)
                self.gradient_weights = cupy.asarray(self.gradient_weights)
                self.laplacian_weights = cupy.asarray(self.laplacian_weights)
            return
        # Just update the link variables
        edges = self.edges
        directions = self.edge_directions
        if self.link_exponents is None:
            link_variables = xp.ones(len(directions))
        else:
            link_variables = xp.exp(
                -1j * xp.einsum("ij, ij -> i", self.link_exponents, directions)
            )
        with warnings.catch_warnings():
            # This is faster than re-creating the sparse matrices from scratch.
            warnings.filterwarnings("ignore", category=sp.SparseEfficiencyWarning)
            # Update gradient for psi
            values = self.gradient_weights * link_variables
            rows = self.gradient_link_rows
            cols = self.gradient_link_cols
            # self.psi_gradient[rows, cols] = values
            _spmatrix_set_many(self.psi_gradient, rows, cols, values)
            # Update Laplacian for psi
            areas = self.areas
            weights = self.laplacian_weights
            values = xp.concatenate(
                [
                    weights * link_variables / areas[edges[:, 0]],
                    weights * link_variables.conjugate() / areas[edges[:, 1]],
                ]
            )
            # Only update rows that are not fixed by boundary conditions
            if self.fix_psi:
                free_rows = self.laplacian_free_rows[: len(self.laplacian_link_rows)]
                rows = self.laplacian_link_rows[free_rows]
                cols = self.laplacian_link_cols[free_rows]
                values = values[free_rows]
            else:
                rows = self.laplacian_link_rows
                cols = self.laplacian_link_cols
            # self.psi_laplacian[rows, cols] = values
            _spmatrix_set_many(self.psi_laplacian, rows, cols, values)

    def get_supercurrent(self, psi: np.ndarray):
        """Compute the supercurrent on the edges.

        Args:
            psi: The value of the complex order parameter.

        Returns:
            The supercurrent at each edge.
        """
        return (psi.conjugate()[self.edges[:, 0]] * (self.psi_gradient @ psi)).imag



from typing import Any, Dict, List, Literal, Optional, Sequence, Tuple, Union

import h5py
import matplotlib.pyplot as plt
import numpy as np

from ..solution.data import get_data_range
from ..solution.solution import Solution
from .common import DEFAULT_QUANTITIES, PLOT_DEFAULTS, Quantity, auto_grid
from .io import get_plot_data, get_state_string


def generate_snapshots(
    input_path: str,
    times: Union[float, Sequence[float]],
    quantities: Union[str, Sequence[str]] = DEFAULT_QUANTITIES,
    shading: Literal["flat", "gouraud"] = "gouraud",
    autoscale: bool = False,
    max_cols: int = 4,
    dimensionless: bool = False,
    xlim: Optional[Tuple[float, float]] = None,
    ylim: Optional[Tuple[float, float]] = None,
    axis_labels: bool = False,
    axes_off: bool = False,
    title_off: bool = False,
    full_title: bool = True,
    figure_kwargs: Optional[Dict[str, Any]] = None,
) -> List[Tuple[plt.Figure, plt.Axes]]:
    """Generates snapshots of a TDGL simulation.

    Args:
        input_path: The path to an H5 file containing
            the :class:`tdgl.Solution` you would like to animate.
        times: The dimensionless time(s) for which to generate a snapshot.
        quantities: The names of the quantities to animate.
        shading: Shading method, "flat" or "gouraud". See matplotlib.pyplot.tripcolor.
        autoscale: Autoscale colorbar limits to exclude outlier points.
        max_cols: The maxiumum number of columns in the subplot grid.
        dimensionless: Use dimensionless units for axes
        xlim: x-axis limits
        ylim: y-axis limits
        axes_off: Turn off the axes for each subplot.
        title_off: Turn off the figure suptitle.
        full_title: Include the full "state" for each frame in the figure suptitle.
        figure_kwargs: Keyword arguments passed to ``plt.subplots()`` when creating
            the figure.

    Returns:
        The matplotlib figure and axes for each time in ``times``
    """
    if isinstance(times, (int, float)):
        times = [times]
    if quantities is None:
        quantities = Quantity.get_keys()
    if isinstance(quantities, str):
        quantities = [quantities]
    quantities = [Quantity.from_key(name.upper()) for name in quantities]
    num_plots = len(quantities)
    figure_kwargs = figure_kwargs or dict()
    figure_kwargs.setdefault("constrained_layout", True)
    default_figsize = (
        3.25 * min(max_cols, num_plots),
        2.5 * max(1, num_plots // max_cols),
    )
    figure_kwargs.setdefault("figsize", default_figsize)
    figure_kwargs.setdefault("sharex", True)
    figure_kwargs.setdefault("sharey", True)

    solution = Solution.from_hdf5(input_path)
    device = solution.device
    mesh = device.mesh
    mesh = device.mesh
    if dimensionless:
        scale = 1
        units_str = "\\xi"
    else:
        scale = device.layer.coherence_length
        units_str = f"{device.ureg(device.length_units).units:~L}"
    x, y = scale * mesh.sites.T

    figures = []

    with h5py.File(input_path, "r") as h5file:
        _, max_step = get_data_range(h5file)
        for time in times:
            step = solution.closest_solve_step(time)

            fig, axes = auto_grid(num_plots, max_cols=max_cols, **figure_kwargs)
            state = get_state_string(h5file, step, max_step)
            if not full_title:
                state = state.split(",")[0]
            if not title_off:
                fig.suptitle(state)

            for quantity, ax in zip(quantities, axes.flat):
                ax: plt.Axes
                opts = PLOT_DEFAULTS[quantity]
                values, _, _ = get_plot_data(h5file, mesh, quantity, step)
                if autoscale:
                    mask = np.abs(values - np.mean(values)) <= 6 * np.std(values)
                else:
                    mask = np.ones_like(values, dtype=bool)
                if opts.vmin is None:
                    vmin = np.min(values[mask])
                else:
                    vmin = opts.vmin
                if opts.vmax is None:
                    vmax = np.max(values[mask])
                else:
                    vmax = opts.vmax
                if opts.symmetric:
                    vmax = max(abs(vmin), abs(vmax))
                    vmin = -vmax
                collection = ax.tripcolor(
                    x,
                    y,
                    values,
                    triangles=mesh.elements,
                    shading=shading,
                    cmap=opts.cmap,
                    vmin=vmin,
                    vmax=vmax,
                )
                cbar = fig.colorbar(collection, ax=ax)
                cbar.set_label(opts.clabel)
                ax.set_aspect("equal")
                ax.set_title(quantity.value)
                ax.set_xlim(xlim)
                ax.set_ylim(ylim)
                if axes_off:
                    ax.axis("off")
                if axis_labels:
                    ax.set_xlabel(f"$x$ [${units_str}$]")
                    ax.set_ylabel(f"$y$ [${units_str}$]")

            figures.append((fig, axes))

    return figures



import argparse
import logging
import os

from .visualization import (
    DEFAULT_QUANTITIES,
    InteractivePlot,
    MultiInteractivePlot,
    Quantity,
    convert_to_xdmf,
    create_animation,
    generate_snapshots,
    monitor_solution,
)

logger = logging.getLogger("visualize")


def make_parser() -> argparse.ArgumentParser:
    quantities_args = ("-q", "--quantities")
    quantities_kwargs = dict(
        type=lambda s: str(s).upper(),
        choices=Quantity.get_keys() + ["ALL"],
        nargs="*",
        help="Name(s) of the quantities to show.",
    )

    parser = argparse.ArgumentParser(description="Visualize TDGL simulation data.")

    # Common arguments
    parser.add_argument("-i", "--input", type=str, help="H5 file to visualize.")
    parser.add_argument("-o", "--output", type=str, help="Output file path.")
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Run in verbose mode.",
    )
    parser.add_argument(
        "--shading",
        type=str,
        choices=["flat", "gouraud"],
        default="gouraud",
        help="Shading method, see matplotlib.pyplot.tripcolor.",
    )
    parser.add_argument(
        "--dimensionless", action="store_true", help="Use dimensionless x-y units."
    )
    parser.add_argument(
        "--xlim", type=float, nargs=2, default=None, help="x-axis limits"
    )
    parser.add_argument(
        "--ylim", type=float, nargs=2, default=None, help="y-axis limits"
    )
    parser.add_argument(
        "--axis-labels", action="store_true", help="Add x-y axis labels."
    )
    parser.add_argument(
        "--autoscale",
        action="store_true",
        help="Autoscale colorbar limits at each frame.",
    )
    parser.add_argument(
        "--axes-off",
        action="store_true",
        help="Turn the axes off.",
    )
    parser.add_argument(
        "--title-off",
        action="store_true",
        help="Turn figure title off.",
    )
    parser.add_argument(
        "--figsize",
        type=float,
        nargs=2,
        default=None,
        help="Figure size (width, height) in inches.",
    )
    parser.add_argument(
        "-d", "--dpi", type=float, default=200, help="Resolution in dots per inch."
    )

    subparsers = parser.add_subparsers()

    # Interactive plot
    interactive_parser = subparsers.add_parser(
        "interactive", help="Create an interactive plot of one or more quantities."
    )
    interactive_parser.add_argument(*quantities_args, **quantities_kwargs)
    interactive_parser.set_defaults(func=visualize_tdgl)

    # Matplotlib animate
    animate_parser = subparsers.add_parser(
        "animate", help="Create an animation of the TDGL data."
    )
    animate_parser.add_argument(
        "-f", "--fps", type=int, default=30, help="Frame rate of the animation."
    )
    animate_parser.add_argument(
        "--min-frame",
        type=int,
        default=0,
        help="The first frame to render.",
    )
    animate_parser.add_argument(
        "--max-frame",
        type=int,
        default=-1,
        help="The last frame to render (-1 indicates the last step in the simulation).",
    )
    animate_parser.add_argument(*quantities_args, **quantities_kwargs)
    animate_parser.set_defaults(func=animate_tdgl)

    # Live monitoring during solve
    monitor_parser = subparsers.add_parser(
        "monitor", help="Visualize the results of a simulation as it is running."
    )
    monitor_parser.add_argument(
        "--interval", type=float, default=1, help="Monitor update interval in seconds."
    )
    monitor_parser.add_argument(*quantities_args, **quantities_kwargs)
    monitor_parser.set_defaults(func=monitor_tdgl)

    # Convert data format
    convert_parser = subparsers.add_parser(
        "convert", help="Convert a Solution from HDF5 to another data format."
    )
    convert_parser.add_argument(
        "--format",
        type=str,
        choices=["xdmf"],
        help="Data format into which to convert the Solution.",
    )
    convert_parser.set_defaults(func=convert_tdgl)

    # Generate snapshots
    snap_parser = subparsers.add_parser(
        "snapshot", help="Generate snapshots of a Solution."
    )
    snap_parser.add_argument(
        "-t",
        "--times",
        type=float,
        nargs="+",
        help="The time(s) at which to generate a snapshot.",
    )
    snap_parser.add_argument(*quantities_args, **quantities_kwargs)
    snap_parser.set_defaults(func=snapshot_tdgl)

    return parser


def animate_tdgl(args: argparse.Namespace) -> None:
    kwargs = dict(
        input_file=args.input,
        output_file=args.output,
        logger=logger,
        shading=args.shading,
        dpi=args.dpi,
        fps=args.fps,
        min_frame=args.min_frame,
        max_frame=args.max_frame,
        autoscale=args.autoscale,
        dimensionless=args.dimensionless,
        xlim=args.xlim,
        ylim=args.ylim,
        axis_labels=args.axis_labels,
        axes_off=args.axes_off,
        title_off=args.title_off,
    )
    if args.figsize is not None:
        kwargs["figure_kwargs"] = dict(figsize=args.figsize)
    if args.quantities is None or "ALL" in args.quantities:
        kwargs["quantities"] = DEFAULT_QUANTITIES
    else:
        kwargs["quantities"] = args.quantities
    create_animation(**kwargs)


def visualize_tdgl(args: argparse.Namespace) -> None:
    kwargs = dict(
        input_file=args.input,
        shading=args.shading,
        dimensionless=args.dimensionless,
        xlim=args.xlim,
        ylim=args.ylim,
        axis_labels=args.axis_labels,
        logger=logger,
    )
    if args.figsize is not None:
        kwargs["figure_kwargs"] = dict(figsize=args.figsize)
    if args.quantities is None:
        InteractivePlot(**kwargs).show()
        return
    if "ALL" not in args.quantities:
        kwargs["quantities"] = args.quantities
    MultiInteractivePlot(**kwargs).show()


def monitor_tdgl(args: argparse.Namespace) -> None:
    dirname = os.path.dirname(args.input)
    fname = os.path.basename(args.input) + ".tmp"
    h5path = os.path.join(dirname, fname)
    kwargs = dict(
        h5path=h5path,
        update_interval=args.interval,
        shading=args.shading,
        xlim=args.xlim,
        ylim=args.ylim,
        autoscale=args.autoscale,
        dimensionless=args.dimensionless,
    )
    if args.figsize is not None:
        kwargs["figure_kwargs"] = dict(figsize=args.figsize)
    if args.quantities is None or "ALL" in args.quantities:
        kwargs["quantities"] = DEFAULT_QUANTITIES
    else:
        kwargs["quantities"] = args.quantities
    monitor_solution(**kwargs)


def convert_tdgl(args: argparse.Namespace) -> None:
    # I may add other convert_funcs if they are requested.
    convert_funcs = {"xdmf": convert_to_xdmf}
    convert_func = convert_funcs[args.format.lower()]
    convert_func(
        path_to_solution=args.input,
        xdmf_path=args.output,
        dimensionless=args.dimensionless,
    )


def snapshot_tdgl(args: argparse.Namespace) -> None:
    if args.output is not None:
        logger.warning("Argument --output is ignored by the snapshot command.")
    dirname = os.path.dirname(args.input)
    basename = os.path.basename(args.input)
    name, _ = os.path.splitext(basename)
    output_file = os.path.join(dirname, f"{name}_[t={{time:.0f}}].png")

    kwargs = dict(
        input_path=args.input,
        times=args.times,
        shading=args.shading,
        autoscale=args.autoscale,
        dimensionless=args.dimensionless,
        xlim=args.xlim,
        ylim=args.ylim,
        axis_labels=args.axis_labels,
        axes_off=args.axes_off,
        title_off=args.title_off,
    )
    if args.figsize is not None:
        kwargs["figure_kwargs"] = dict(figsize=args.figsize)
    if args.quantities is None or "ALL" in args.quantities:
        kwargs["quantities"] = DEFAULT_QUANTITIES
    else:
        kwargs["quantities"] = args.quantities

    figures = generate_snapshots(**kwargs)
    for time, (fig, axes) in zip(args.times, figures):
        fig.savefig(output_file.format(time=time), dpi=args.dpi, bbox_inches="tight")


def main(args: argparse.Namespace) -> None:
    logger.setLevel(logging.DEBUG if args.verbose else logging.INFO)
    args.func(args)


if __name__ == "__main__":
    main(make_parser().parse_args())



from ..parameter import Parameter


def linear_ramp(x, y, z, *, t, tmin, tmax, initial: float = 0.0, final: float = 1.0):
    """A linear ramp from ``initial`` to ``final`` over time period ``tmin`` to ``tmax``.

    For times less than ``tmin``, the value stays at ``initial``. For times greater
    than ``tmax``, the value stays at ``final``.
    """
    if t < tmin:
        return initial
    elif t < tmax:
        return initial + (final - initial) * (t - tmin) / (tmax - tmin)
    return final


def LinearRamp(*, tmin: float, tmax, initial: float = 0.0, final: float = 1.0):
    """A linear ramp from ``initial`` to ``final`` over time period ``tmin`` to ``tmax``.

    For times less than ``tmin``, the value stays at ``initial``. For times greater
    than ``tmax``, the value stays at ``final``.

    Args:
        tmin: Ramp start time
        tmax: Ramp end time
        initial: Ramp initial value
        final: Ramp final value

    Returns:
        A :class:`tdgl.Parameter` that produces a linear ramp.
    """
    return Parameter(
        linear_ramp,
        tmin=tmin,
        tmax=tmax,
        initial=initial,
        final=final,
        time_dependent=True,
    )


def Scale(func, **kwargs):
    """An arbitrary time-dependent scale factor.

    Args:
        func: A callable with signature func(x, y, z, *, t, **kwargs) that
            produces a time-dependent scale factor.

    Returns:
        A :class:`tdgl.Parameter` that evaluates ``func``
    """
    kwargs["time_dependent"] = True
    return Parameter(func, **kwargs)
